"""
This special modules is for variations that do easily fit into othet categories
"""
import exponax as ex
import jax.numpy as jnp

from ..._base_scenario import BaseScenario


class UnbalancedAdvection(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.1

    advection_coef_vector: tuple[float, ...] = (
        0.01,
        -0.04,
        0.005,
    )  # Needs to be as long as num_spatial_dims

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError("Unbalanced advection is only supported for 2D and 3D")
        if len(self.advection_coef_vector) != self.num_spatial_dims:
            raise ValueError(
                "Advection coefficient vector needs to be as long as the number of spatial dimensions"
            )

    def get_ref_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Advection(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            velocity=jnp.array(self.advection_coef_vector),
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Advection(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            velocity=jnp.array(self.advection_coef_vector),
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_unbal_adv"


class DiagonalDiffusion(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.1

    diffusion_coef_vector: tuple[float, ...] = (
        0.001,
        0.002,
    )  # Needs to be as long as num_spatial_dims

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError("Diagonal diffusion is only supported for 2D and 3D")
        if len(self.diffusion_coef_vector) != self.num_spatial_dims:
            raise ValueError(
                "Diffusion coefficient vector needs to be as long as the number of spatial dimensions"
            )

    def get_ref_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Diffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            diffusivity=jnp.array(self.diffusion_coef_vector),
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Diffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            diffusivity=jnp.array(self.diffusion_coef_vector),
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_diag_diff"


class AnisotropicDiffusion(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.1

    diffusion_coef_matrix: tuple[tuple[float, ...], ...] = (
        (0.001, 0.0005),
        (0.0005, 0.002),
    )
    """
    Needs to be a square matrix with the same size as the number of spatial
    dimensions, given as a tuple of tuples.

    Also has to be symmetric and positive definite.
    """

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError("Anisotropic diffusion is only supported for 2D and 3D")
        if len(self.diffusion_coef_matrix) != self.num_spatial_dims:
            raise ValueError(
                "Diffusion coefficient matrix needs to be as long as the number of spatial dimensions"
            )
        for row in self.diffusion_coef_matrix:
            if len(row) != self.num_spatial_dims:
                raise ValueError("Diffusion coefficient matrix needs to be square")

        # No check for SPD for now

    def get_ref_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Diffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            diffusivity=jnp.array(self.diffusion_coef_matrix),
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Diffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            diffusivity=jnp.array(self.diffusion_coef_matrix),
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_aniso_diff"


class SpatiallyMixedDispersion(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.001
    dispersion_coef: float = 0.00025

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError(
                "Spatially mixed dispersion is only supported for 2D and 3D"
            )

    def get_ref_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Dispersion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            dispersivity=self.dispersion_coef,
            advect_on_diffusion=True,
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.Dispersion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            dispersivity=self.dispersion_coef,
            advect_on_diffusion=True,
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_mix_disp"


class SpatiallyMixedHyperDiffusion(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.00001
    hyp_diffusion_coef: float = -0.000075

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError(
                "Spatially mixed hyperdiffusion is only supported for 2D and 3D"
            )

    def get_ref_stepper(self) -> ex.BaseStepper:
        return ex.stepper.HyperDiffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            # Need minus to match the sign of the diffusion coefficient
            hyper_diffusivity=-self.hyp_diffusion_coef,
            diffuse_on_diffuse=True,
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.HyperDiffusion(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            # Need minus to match the sign of the diffusion coefficient
            hyper_diffusivity=-self.hyp_diffusion_coef,
            diffuse_on_diffuse=True,
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_mix_hyp"



import exponax as ex

from ..._base_scenario import BaseScenario


class Linear(BaseScenario):
    alphas: tuple[float, ...] = (0.0, -0.025, 0.0, 0.0, 0.0)
    coarse_proportion: float = 0.5

    def get_ref_stepper(self):
        return ex.stepper.generic.NormalizedLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            normalized_linear_coefficients=self.alphas,
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.generic.NormalizedLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            normalized_linear_coefficients=tuple(
                f * self.coarse_proportion for f in self.alphas
            ),
        )

    def get_scenario_name(self) -> str:
        active_indices = []
        for i, a in enumerate(self.alphas):
            if a != 0.0:
                active_indices.append(i)
        return f"{self.num_spatial_dims}d_norm_lin_{'_'.join(str(i) for i in active_indices)}"


class LinearSimple(Linear):
    linear_alpha: float = -0.025
    linear_term_order: int = 1

    def __post_init__(self):
        self.alphas = (0.0,) * self.linear_term_order + (self.linear_alpha,)


class Advection(Linear):
    advection_alpha: float = -0.025

    def __post_init__(self):
        self.alphas = (0.0, self.advection_alpha, 0.0, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_adv"


class Diffusion(Linear):
    diffusion_alpha: float = 8e-4

    def __post_init__(self):
        self.alphas = (0.0, 0.0, self.diffusion_alpha, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_diff"


class AdvectionDiffusion(Linear):
    advection_alpha: float = -0.025
    diffusion_alpha: float = 8e-4

    def __post_init__(self):
        self.alphas = (0.0, self.advection_alpha, self.diffusion_alpha)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_adv_diff"


class Dispersion(Linear):
    dispersion_alpha: float = 2.5e-7

    def __post_init__(self):
        self.alphas = (0.0, 0.0, 0.0, self.dispersion_alpha, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_disp"


class HyperDiffusion(Linear):
    hyp_diffusion_alpha: float = -7.5e-10

    def __post_init__(self):
        self.alphas = (0.0, 0.0, 0.0, 0.0, self.hyp_diffusion_alpha)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_hyp_diff"


class FirstFour(Linear):
    advection_alpha: float = -0.025
    diffusion_alpha: float = 8e-4
    dispersion_alpha: float = 2.5e-7
    hyp_diffusion_alpha: float = -7.5e-10

    def __post_init__(self):
        self.alphas = (
            0.0,
            self.advection_alpha,
            self.diffusion_alpha,
            self.dispersion_alpha,
            self.hyp_diffusion_alpha,
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_four"



import apebench


def test_trainer_parsing():
    advection_scenario = apebench.scenarios.difficulty.Advection()

    trainer = advection_scenario.get_trainer(
        train_config="one",
    )

    assert trainer.loss_configuration.num_rollout_steps == 1
    del trainer

    trainer = advection_scenario.get_trainer(train_config="sup;3")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is False
    del trainer

    trainer = advection_scenario.get_trainer(train_config="sup;3;True")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is True
    del trainer

    trainer = advection_scenario.get_trainer(train_config="sup;3;False")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is False
    del trainer

    trainer = advection_scenario.get_trainer(train_config="div;3")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is False
    assert trainer.loss_configuration.cut_div_chain is False
    del trainer

    trainer = advection_scenario.get_trainer(train_config="div;3;True")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is True
    assert trainer.loss_configuration.cut_div_chain is False
    del trainer

    trainer = advection_scenario.get_trainer(train_config="div;3;False")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is False
    assert trainer.loss_configuration.cut_div_chain is False
    del trainer

    trainer = advection_scenario.get_trainer(train_config="div;3;True;True")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is True
    assert trainer.loss_configuration.cut_div_chain is True
    del trainer

    trainer = advection_scenario.get_trainer(train_config="div;3;False;True")

    assert trainer.loss_configuration.num_rollout_steps == 3
    assert trainer.loss_configuration.cut_bptt is False
    assert trainer.loss_configuration.cut_div_chain is True
    del trainer



import exponax as ex

from ..._base_scenario import BaseScenario


class Nonlinear(BaseScenario):
    """
    Uses the single channel convection mode to not have channels grow with
    spatial dimensions.
    """

    alphas: tuple[float, ...] = (0.0, 0.0, 0.00003, 0.0, 0.0)
    betas: tuple[float, float, float] = (0.0, -0.0125, 0.0)

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: int = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        pass

    def _build_stepper(self, alphas, betas):
        substepped_alphas = tuple(a / self.num_substeps for a in alphas)
        substepped_betas = tuple(b / self.num_substeps for b in betas)

        substepped_stepper = ex.stepper.generic.NormalizedNonlinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            normalized_linear_coefficients=substepped_alphas,
            normalized_nonlinear_coefficients=substepped_betas,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.alphas, self.betas)

    def get_coarse_stepper(self):
        return self._build_stepper(
            tuple(f * self.coarse_proportion for f in self.alphas),
            tuple(f * self.coarse_proportion for f in self.betas),
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_nonlin"


class BurgersSingleChannel(Nonlinear):
    convection_sc_beta: float = -0.0125
    diffusion_alpha: float = 0.00003

    def __post_init__(self):
        self.alphas = (0.0, 0.0, self.diffusion_alpha, 0.0, 0.0)
        self.betas = (0.0, self.convection_sc_beta, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_burgers_sc"


class KortewegDeVries(Nonlinear):
    convection_sc_beta: float = -0.0125
    dispersion_alpha: float = -8.5e-7
    hyp_diffusion_alpha: float = -2e-9

    def __post_init__(self):
        self.alphas = (0.0, 0.0, 0.0, self.dispersion_alpha, self.hyp_diffusion_alpha)
        self.betas = (0.0, self.convection_sc_beta, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_kdv"


class KuramotoSivashinsky(Nonlinear):
    gradient_norm_beta: float = -0.00025
    diffusion_alpha: float = -0.000025
    hyp_diffusion_alpha: float = -3.0e-9

    num_warmup_steps: int = 500  # Overwrite
    vlim: tuple[float, float] = (-6.5, 6.5)  # Overwrite

    report_metrics: str = "mean_nRMSE,mean_correlation"  # Overwrite

    def __post_init__(self):
        self.alphas = (0.0, 0.0, self.diffusion_alpha, 0.0, self.hyp_diffusion_alpha)
        self.betas = (0.0, 0.0, self.gradient_norm_beta)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_ks"


class FisherKPP(Nonlinear):
    quadratic_beta: float = -0.02
    drag_alpha: float = 0.02
    diffusion_alpha: float = 4e-6

    ic_config: str = "clamp;0.0;1.0;fourier;5;false;false"  # Overwrite

    def __post_init__(self):
        self.alphas = (self.drag_alpha, 0.0, self.diffusion_alpha, 0.0, 0.0)
        self.betas = (self.quadratic_beta, 0.0, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_fisher"



"""
Scenarios using the general nonlinear interface
"""

import exponax as ex

from ..._base_scenario import BaseScenario


class Nonlinear(BaseScenario):
    """
    Uses the single channel convection mode to not have channels grow with
    spatial dimensions.

    By default single-channel Burgers
    """

    gammas: tuple[float, ...] = (0.0, 0.0, 1.5, 0.0, 0.0)
    deltas: tuple[float, float, float] = (0.0, -1.5, 0.0)

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: int = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        pass

    def _build_stepper(self, gammas, deltas):
        substepped_gammas = tuple(g / self.num_substeps for g in gammas)
        substepped_deltas = tuple(d / self.num_substeps for d in deltas)

        substepped_stepper = ex.stepper.generic.DifficultyNonlinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            linear_difficulties=substepped_gammas,
            nonlinear_difficulties=substepped_deltas,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.gammas, self.deltas)

    def get_coarse_stepper(self):
        return self._build_stepper(
            tuple(f * self.coarse_proportion for f in self.gammas),
            tuple(f * self.coarse_proportion for f in self.deltas),
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_nonlin"


class BurgersSingleChannel(Nonlinear):
    convection_sc_delta: float = -1.5
    diffusion_gamma: float = 1.5

    def __post_init__(self):
        self.gammas = (0.0, 0.0, self.diffusion_gamma, 0.0, 0.0)
        self.deltas = (0.0, self.convection_sc_delta, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_burgers_sc"


class KortewegDeVries(Nonlinear):
    convection_sc_delta: float = -2.0
    dispersion_gamma: float = -14.0
    hyp_diffusion_gamma: float = -9.0

    def __post_init__(self):
        self.gammas = (0.0, 0.0, 0.0, self.dispersion_gamma, self.hyp_diffusion_gamma)
        self.deltas = (0.0, self.convection_sc_delta, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_kdv"


class KuramotoSivashinsky(Nonlinear):
    gradient_norm_delta: float = -6.0
    diffusion_gamma: float = -1.2  # Negative diffusion! producing energy
    hyp_diffusion_gamma: float = -15.0

    num_warmup_steps: int = 500  # Overwrite
    vlim: tuple[float, float] = (-6.5, 6.5)  # Overwrite

    report_metrics: str = "mean_nRMSE,mean_correlation"  # Overwrite

    def __post_init__(self):
        self.gammas = (0.0, 0.0, self.diffusion_gamma, 0.0, self.hyp_diffusion_gamma)
        self.deltas = (0.0, 0.0, self.gradient_norm_delta)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_ks"


class FisherKPP(Nonlinear):
    quadratic_delta: float = -0.02
    drag_gamma: float = 0.02
    diffusion_gamma: float = 0.2

    ic_config: str = "clamp;0.0;1.0;fourier;5;false;false"  # Overwrite

    def __post_init__(self):
        self.gammas = (self.drag_gamma, 0.0, self.diffusion_gamma)
        self.deltas = (self.quadratic_delta, 0.0, 0.0)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_fisher"



import os
import pathlib
from typing import Optional, Union

import equinox as eqx
import pandas as pd
from tqdm.autonotebook import tqdm

from ._utils import melt_loss, melt_metrics, melt_sample_rollouts, read_in_kwargs
from .scenarios import scenario_dict


def run_experiment(
    *,
    scenario: str,
    task: str,
    net: str,
    train: str,
    start_seed: int,
    num_seeds: int,
    **scenario_kwargs,
) -> tuple[pd.DataFrame, eqx.Module]:
    """
    Execute a single experiment.

    Only accepts keyword arguments, requires some main arguments and additional
    arguments that can depend on the chosen scenario.

    **Arguments:**

    * `scenario`: The scenario to run, must be a key in
        `apebench.scenarios.scenario_dict`.
    * `task`: The task to run, can be `"predict"` or `"correct;XX"`
        where `"XX"` is the correction mode, e.g., `"correct;sequential"`.
    * `net`: The network to use, must be a compatible descriptor of a
        network architecture, e.g., `"Conv;34;10;relu"`.
    * `train`: The training methodology to use, i.e., how reference
        solver and emulator interact during training. One-step supervised
        training is achieved by `"one"`.
    * `start_seed`: The integer at which the list of seeds starts from.
    * `num_seeds`: The number of seeds to run (in parallel). For many 1D
        scenarios at realistic resolutions (`num_points` ~ 200), doing ten seeds
        in parallel is usually fine for modern GPUs. For scenarios in 2D and 3D
        at realistic resolutions, this likely has to be set to 1 and seed
        processing must be done sequentially via `run_study`.

    **Returns:**

    * `data`: The `pandas.DataFrame` containing the results of the
        experiment. Will contain the columns `seed`, `scenario`, `task`, `net`,
        `train`, `scenario_kwargs`, the metrics, losses and sample rollouts. Can
        be further post-processed, e.g., via `apebench.melt_metrics`.
    * `trained_neural_stepper_s`: Equinox modules containing the
        trained neural emulators. Note that if `num_seeds` is greater than 1,
        weight arrays have a leading (seed-)batch axis.
    """
    scenario = scenario_dict[scenario](**scenario_kwargs)

    data, trained_neural_stepper_s = scenario(
        task_config=task,
        network_config=net,
        train_config=train,
        start_seed=start_seed,
        num_seeds=num_seeds,
    )

    if len(scenario_kwargs) == 0:
        data["scenario_kwargs"] = "{}"
    else:
        data["scenario_kwargs"] = str(scenario_kwargs)

    return data, trained_neural_stepper_s


def get_experiment_name(
    *,
    scenario: str,
    task: str,
    net: str,
    train: str,
    start_seed: int,
    num_seeds: int,
    **scenario_kwargs,
) -> str:
    """
    Produce a unique name for an experiment.
    """
    additional_infos = []
    for key, value in scenario_kwargs.items():
        additional_infos.append(f"{key}={value}")
    if len(additional_infos) > 0:
        additional_infos = ",".join(additional_infos)
        additional_infos = f"__{additional_infos}__"
    else:
        additional_infos = "__"

    end_seed = start_seed + num_seeds
    experiment_name = f"{scenario}{additional_infos}{task}__{net}__{train}__{start_seed}-{end_seed - 1}"
    return experiment_name


def run_study(
    configs: list[dict],
    base_path: str,
    *,
    overwrite: bool = False,
) -> tuple[list[pathlib.Path], list[pathlib.Path]]:
    """
    Execute a study with multiple experiments.

    By default skips experiments that have already been conducted.

    **Arguments:**

    * `configs`: A list of dictionaries, each containing the
        keyword arguments for [`apebench.run_experiment`][].
    * `base_path`: The base path to store the results in.
    * `overwrite`: Whether to overwrite existing results.

    **Returns:**

    * `raw_file_list`: A list of paths to the raw data files.
    * `network_weights_list`: A list of paths to the
        network weights.
    """
    raw_file_list = []
    network_weights_list = []

    for config in configs:
        experiment_name = get_experiment_name(**config)

        print("Considering")
        print(experiment_name)

        raw_data_folder = base_path / pathlib.Path("raw")
        os.makedirs(raw_data_folder, exist_ok=True)
        raw_data_path = raw_data_folder / pathlib.Path(f"{experiment_name}.csv")

        network_weights_folder = base_path / pathlib.Path("network_weights")
        os.makedirs(network_weights_folder, exist_ok=True)
        network_weights_path = network_weights_folder / pathlib.Path(
            f"{experiment_name}.eqx"
        )

        raw_file_list.append(raw_data_path)
        network_weights_list.append(network_weights_path)

        if (
            os.path.exists(raw_data_path)
            and os.path.exists(network_weights_path)
            and not overwrite
        ):
            print("Skipping, already trained ...")
            print()
            continue

        data, trained_neural_stepper_s = run_experiment(**config)

        data.to_csv(raw_data_path)
        eqx.tree_serialise_leaves(
            network_weights_path,
            trained_neural_stepper_s,
        )

        del data
        del trained_neural_stepper_s

        print("Finished training!")
        print()

    return raw_file_list, network_weights_list


def melt_concat_metrics_from_list(
    raw_file_list: list[pathlib.Path],
    *,
    metric_name: Union[str, list[str]] = "mean_nRMSE",
) -> pd.DataFrame:
    """
    Melt and concatenate metrics from a list of raw files. Use this function on
    the results of [`apebench.run_study`][].

    **Arguments:**

    * `raw_file_list`: A list of paths to the raw data
      files.
    * `metric_name`: The name of the metric to melt.

    **Returns:**

    * `metric_df`: The DataFrame containing the metrics.
    """
    metric_df_s = []
    for file_name in tqdm(
        raw_file_list,
        desc="Melt and Concat metrics",
    ):
        data = pd.read_csv(file_name)
        data = melt_metrics(data, metric_name=metric_name)
        metric_df_s.append(data)

    metric_df = pd.concat(metric_df_s)

    return metric_df


def melt_concat_loss_from_list(
    raw_file_list: list[pathlib.Path],
) -> pd.DataFrame:
    """
    Melt and concatenate loss from a list of raw files. Use this function on the
    results of [`apebench.run_study`][].

    **Arguments:**

    * `raw_file_list`: A list of paths to the raw data files.

    **Returns:**

    * `loss_df`: The DataFrame containing the loss.
    """
    loss_df_s = []
    for file_name in tqdm(
        raw_file_list,
        desc="Melt and Concat loss",
    ):
        data = pd.read_csv(file_name)
        data = melt_loss(data)
        loss_df_s.append(data)

    loss_df = pd.concat(loss_df_s)

    return loss_df


def melt_concat_sample_rollouts_from_list(
    raw_file_list: list[pathlib.Path],
) -> pd.DataFrame:
    """
    Melt and concatenate sample rollouts from a list of raw files. Use this
    function on the results of [`apebench.run_study`][].

    **Arguments:**

    * `raw_file_list`: A list of paths to the raw data files.

    **Returns:**

    * `sample_rollout_df`: The DataFrame containing the sample rollouts.
    """
    sample_rollout_df_s = []
    for file_name in tqdm(
        raw_file_list,
        desc="Melt and Concat sample rollouts",
    ):
        data = pd.read_csv(file_name)
        data = melt_sample_rollouts(data)
        sample_rollout_df_s.append(data)

    sample_rollout_df = pd.concat(sample_rollout_df_s)

    return sample_rollout_df


def melt_concat_from_list(
    raw_file_list: list[pathlib.Path],
    base_path: str,
    *,
    metric_name: Union[str, list[str]] = "mean_nRMSE",
    metric_file_name: str = "metrics",
    loss_file_name: str = "train_loss",
    sample_rollout_file_name: str = "sample_rollout",
    do_metrics: bool = True,
    do_loss: bool = False,
    do_sample_rollouts: bool = False,
) -> tuple[Optional[pathlib.Path], Optional[pathlib.Path], Optional[pathlib.Path]]:
    """
    Melt and concatenate metrics, loss and sample rollouts from a list of raw
    files and save the resulting DataFrames to disk as CSV files. Use this
    function on the results of [`apebench.run_study`][].

    **Arguments:**

    * `raw_file_list`: A list of paths to the raw data files.
    * `base_path`: The base path to store the results in.
    * `metric_name`: The name of the metric to melt.
    * `metric_file_name`: The name of the file to save the metrics to.
    * `loss_file_name`: The name of the file to save the loss to.
    * `sample_rollout_file_name`: The name of the file to save the sample
        rollouts to.
    * `do_metrics`: Whether to melt and save the metrics.
    * `do_loss`: Whether to melt and save the loss.
    * `do_sample_rollouts`: Whether to melt and save the sample rollouts.

    **Returns:**

    * `metric_df_file_name`: The path to the metrics file.
    * `loss_df_file_name`: The path to the loss file.
    * `sample_rollout_df_file_name`: The path to the sample rollouts file.
    """
    if do_metrics:
        metric_df = melt_concat_metrics_from_list(
            raw_file_list,
            metric_name=metric_name,
        )
        metric_df_file_name = base_path / pathlib.Path(f"{metric_file_name}.csv")
        metric_df.to_csv(
            metric_df_file_name,
            index=False,
        )
    else:
        metric_df_file_name = None

    if do_loss:
        loss_df = melt_concat_loss_from_list(raw_file_list)
        loss_df_file_name = base_path / pathlib.Path(f"{loss_file_name}.csv")
        loss_df.to_csv(
            loss_df_file_name,
            index=False,
        )
    else:
        loss_df_file_name = None

    if do_sample_rollouts:
        sample_rollout_df = melt_concat_sample_rollouts_from_list(raw_file_list)
        sample_rollout_df_file_name = base_path / pathlib.Path(
            f"{sample_rollout_file_name}.csv"
        )
        sample_rollout_df.to_csv(
            sample_rollout_df_file_name,
            index=False,
        )
    else:
        sample_rollout_df_file_name = None

    return metric_df_file_name, loss_df_file_name, sample_rollout_df_file_name


def run_study_convenience(
    configs: list[dict],
    base_path: Optional[str] = None,
    *,
    overwrite: bool = False,
    metric_name: Union[str, list[str]] = "mean_nRMSE",
    do_metrics: bool = True,
    do_loss: bool = False,
    do_sample_rollouts: bool = False,
    parse_kwargs: bool = True,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, list[pathlib.Path]]:
    """
    Run a study with multiple experiments and melt and concatenate the results.

    **Arguments:**

    * `configs`: A list of dictionaries, each containing the
        keyword arguments for `run_experiment`.
    * `base_path`: The base path to store the results in. If
        `None`, a path is generated based on the hash of the `configs`.
    * `overwrite`: Whether to overwrite existing results.
    * `metric_name`: The name of the metric to melt.
    * `do_metrics`: Whether to melt and save the metrics.
    * `do_loss`: Whether to melt and save the loss.
    * `do_sample_rollouts`: Whether to melt and save the sample rollouts.
    * `parse_kwargs`: Whether to parse the scenario kwargs.

    **Returns:**

    * `metric_df`: The DataFrame containing the metrics.
    * `loss_df`: The DataFrame containing the loss.
    * `sample_rollout_df`: The DataFrame containing the sample rollouts.
    * `network_weights_list`: A list of paths to the network weights.
    """
    if base_path is None:
        config_hash = hash(str(configs))
        base_path = pathlib.Path(f"_results_{config_hash}")

    raw_file_list, network_weights_list = run_study(
        configs,
        base_path,
        overwrite=overwrite,
    )

    melt_concat_from_list(
        raw_file_list,
        base_path,
        metric_name=metric_name,
        do_metrics=do_metrics,
        do_loss=do_loss,
        do_sample_rollouts=do_sample_rollouts,
    )

    if do_metrics:
        metric_df = pd.read_csv(base_path / pathlib.Path("metrics.csv"))
        if parse_kwargs:
            metric_df = read_in_kwargs(metric_df)
    else:
        metric_df = pd.DataFrame()

    if do_loss:
        loss_df = pd.read_csv(base_path / pathlib.Path("train_loss.csv"))
        if parse_kwargs:
            loss_df = read_in_kwargs(loss_df)
    else:
        loss_df = pd.DataFrame()

    if do_sample_rollouts:
        sample_rollout_df = pd.read_csv(base_path / pathlib.Path("sample_rollout.csv"))
        if parse_kwargs:
            sample_rollout_df = read_in_kwargs(sample_rollout_df)
    else:
        sample_rollout_df = pd.DataFrame()

    return metric_df, loss_df, sample_rollout_df, network_weights_list



import exponax as ex

from ..._base_scenario import BaseScenario


class Convection(BaseScenario):
    alphas: tuple[float, ...] = (0.0, 0.0, 3.0e-5, 0.0, 0.0)
    convection_beta: float = -1.25e-2
    conservative: bool = True

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: float = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        self.num_channels = self.num_spatial_dims  # Overwrite

    def _build_stepper(self, convection, alphas):
        substepped_convection = convection / self.num_substeps
        substepped_alphas = tuple(a / self.num_substeps for a in alphas)

        substepped_stepper = ex.stepper.generic.NormalizedConvectionStepper(
            self.num_spatial_dims,
            self.num_points,
            normalized_linear_coefficients=substepped_alphas,
            # Need minus to move the convection to the right hand side
            normalized_convection_scale=-substepped_convection,
            conservative=self.conservative,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.convection_beta, self.alphas)

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(
            self.coarse_proportion * self.convection_beta,
            tuple(f * self.coarse_proportion for f in self.alphas),
        )

    def get_scenario_name(self) -> str:
        active_indices = []
        for i, a in enumerate(self.alphas):
            if a != 0.0:
                active_indices.append(i)
        return f"{self.num_spatial_dims}d_norm_conv_{'_'.join(str(i) for i in active_indices)}"


class Burgers(Convection):
    convection_beta: float = -1.25e-2  # Overwrite
    diffusion_alpha: float = 3.0e-5

    def __post_init__(self):
        self.alphas = (0.0, 0.0, self.diffusion_alpha, 0.0, 0.0)
        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_burgers"


class KuramotoSivashinskyConservative(Convection):
    convection_delta: float = -6.0e-3
    diffusion_alpha: float = -4.0e-5
    hyp_diffusion_alpha: float = -3.0e-8

    num_warmup_steps: int = 500  # Overwrite
    vlim: tuple[float, float] = (-2.5, 2.5)  # Overwrite

    report_metrics: str = "mean_nRMSE,mean_correlation"  # Overwrite

    def __post_init__(self):
        if self.num_spatial_dims != 1:
            raise ValueError(
                "Conservative Kuramoto-Sivashinsky is only defined for 1 spatial dimension. Check out the non-conservative version for 2d."
            )
        self.alphas = (0.0, 0.0, self.diffusion_alpha, 0.0, self.hyp_diffusion_alpha)
        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_norm_ks_cons"



import jax.numpy as jnp

import apebench


def test_ic_set_is_equal_to_first_snapshot():
    # Use a scenario with a warmup
    ks_scenario = apebench.scenarios.difficulty.KuramotoSivashinsky()

    # Train data
    train_data = ks_scenario.get_train_data()
    train_ic_set = ks_scenario.get_train_ic_set()

    assert jnp.array_equal(train_data[:, 0], train_ic_set)

    # Test data
    test_data = ks_scenario.get_test_data()
    test_ic_set = ks_scenario.get_test_ic_set()

    assert jnp.array_equal(test_data[:, 0], test_ic_set)

    # Coarse train data
    coarse_train_data = ks_scenario.get_train_data_coarse()
    coarse_train_ic_set = ks_scenario.get_train_ic_set_coarse()

    assert jnp.array_equal(coarse_train_data[:, 0], coarse_train_ic_set)

    # Coarse test data
    coarse_test_data = ks_scenario.get_test_data_coarse()
    coarse_test_ic_set = ks_scenario.get_test_ic_set_coarse()

    assert jnp.array_equal(coarse_test_data[:, 0], coarse_test_ic_set)



import exponax as ex

from ..._base_scenario import BaseScenario


class Poisson(BaseScenario):
    domain_extent: float = 10.0
    order: int = 2

    num_warmup_steps: int = 0  # Overwrite
    train_temporal_horizon: int = 1  # Overwrite
    test_temporal_horizon: int = 1  # Overwrite

    def __post_init__(self):
        if self.train_temporal_horizon != 1:
            raise ValueError("Only temporal horizon 1 is supported for Poisson")
        if self.test_temporal_horizon != 1:
            raise ValueError("Only temporal horizon 1 is supported for Poisson")
        if self.num_warmup_steps != 0:
            raise ValueError("Warmup steps are not supported for Poisson")

    def get_ref_stepper(self):
        return ex.poisson.Poisson(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            order=self.order,
        )

    def get_coarse_stepper(self):
        raise NotImplementedError("Coarse stepper is not implemented for Poisson")

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_poisson_order{self.order}"



from abc import ABC, abstractmethod
from typing import Callable, Optional, Union

import equinox as eqx
import exponax as ex
import jax
import jax.numpy as jnp
import optax
import pandas as pd
import pdequinox as pdeqx
import trainax as tx
from exponax import BaseStepper
from exponax.ic import BaseRandomICGenerator
from jaxtyping import Array, Float, PRNGKeyArray

from ._corrected_stepper import CorrectedStepper
from .components import (
    activation_fn_dict,
    architecture_dict,
    ic_dict,
    lr_scheduler_dict,
    metric_dict,
    optimizer_dict,
)


class BaseScenario(eqx.Module, ABC):
    # Setting up the discretization
    num_spatial_dims: int = 1
    num_points: int = 160

    # Abstract information about the problem
    num_channels: int = 1

    # Settings for both training and testing
    ic_config: str = "fourier;5;true;true"
    num_warmup_steps: int = 0

    # Setting up the training
    num_train_samples: int = 50
    train_temporal_horizon: int = 50
    train_seed: int = 0

    # For testing
    num_test_samples: int = 30
    test_temporal_horizon: int = 200
    test_seed: int = 773

    # For the training configuration
    optim_config: str = "adam;10_000;warmup_cosine;0.0;1e-3;2_000"
    batch_size: int = 20

    # Information for inspection
    num_trjs_returned: int = 1
    record_loss_every: int = 100
    vlim: tuple[float, float] = (-1.0, 1.0)
    report_metrics: str = "mean_nRMSE"  # separate by commas ","
    callbacks: str = ""  # separate by commas ","

    def get_ic_generator(self) -> BaseRandomICGenerator:
        """
        Overwrite for custom initial condition generation.

        Uses the `ic_config` attribute to determine the type of initial
        condition generation.

        Allows for the following options:

        - `fourier;CUTOFF;ZERO_MEAN;MAX_ONE` for a truncated Fourier series
            with CUTOFF (int) number of modes, ZERO_MEAN (bool) for zero mean,
            and MAX_ONE (bool) for having the initial condition being at max in
            (-1, 1) but not clamped to it
        - `diffused;INTENSITY;ZERO_MEAN;MAX_ONE` for a diffused noise with
            INTENSITY (float) for the intensity, ZERO_MEAN (bool) for zero mean,
            and MAX_ONE (bool) for having the initial condition being at max in
            (-1, 1) but not clamped to it
        - `grf;POWERLAW_EXPONENT;ZERO_MEAN;MAX_ONE` for a Gaussian random
            field with POWERLAW_EXPONENT (float) for the powerlaw exponent,
            ZERO_MEAN (bool) for zero mean, and MAX_ONE (bool) for having the
            initial condition being at max in (-1, 1) but not clamped to it
        - `clamp;LOWER_BOUND;UPPER_BOUND;CONFIG` for clamping the
            configuration to the range of LOWER_BOUND (float) to UPPER_BOUND
            (float) and then using the configuration CONFIG for the generation
            of the initial condition
        """

        def _get_single_channel(config):
            ic_name = config.split(";")[0]
            ic_gen = ic_dict[ic_name.lower()](config, self.num_spatial_dims)
            return ic_gen

        ic_args = self.ic_config.split(";")
        if ic_args[0].lower() == "clamp":
            lower_bound = float(ic_args[1])
            upper_bound = float(ic_args[2])

            ic_gen = _get_single_channel(";".join(ic_args[3:]))
            ic_gen = ex.ic.ClampingICGenerator(
                ic_gen,
                limits=(lower_bound, upper_bound),
            )
        else:
            ic_gen = _get_single_channel(self.ic_config)

        multi_channel_ic_gen = ex.ic.RandomMultiChannelICGenerator(
            [
                ic_gen,
            ]
            * self.num_channels
        )

        return multi_channel_ic_gen

    @abstractmethod
    def get_ref_stepper(self) -> BaseStepper:
        """
        Produces the reference stepper for the scenario.
        """
        pass

    @abstractmethod
    def get_coarse_stepper(self) -> BaseStepper:
        """
        Produces the coarse stepper for the scenario.
        """
        pass

    @abstractmethod
    def get_scenario_name(self) -> str:
        """
        Produces a unique identifier for this scenario
        """
        pass

    @property
    def num_training_steps(self):
        optim_args = self.optim_config.split(";")
        return int(optim_args[1])

    def get_optimizer(self) -> optax.GradientTransformation:
        """
        Returns the optimizer used in the scenario.
        """
        optim_args = self.optim_config.split(";")
        optimizer_name = optim_args[0]
        num_training_steps = int(optim_args[1])
        scheduler_args = optim_args[2:]
        scheduler_name = scheduler_args[0]

        lr_scheduler = lr_scheduler_dict[scheduler_name.lower()](
            ";".join(scheduler_args), num_training_steps
        )
        optimizer = optimizer_dict[optimizer_name.lower()](self.optim_config)(
            lr_scheduler
        )

        return optimizer

    def _produce_ic_set(
        self,
        *,
        stepper: BaseStepper,
        num_samples: int,
        key: PRNGKeyArray,
    ) -> Float[Array, "num_samples num_channels *num_points"]:
        """
        Generate the number of initial conditions as samples requested and
        discretize them on the grid. Requires the `stepper` to warmup the
        initial conditions if necessary.
        """
        ic_distribution = self.get_ic_generator()
        ic_set = ex.build_ic_set(
            ic_distribution,
            num_points=self.num_points,
            num_samples=num_samples,
            key=key,
        )
        if self.num_warmup_steps > 0:
            ic_set = jax.vmap(
                ex.repeat(
                    stepper,
                    self.num_warmup_steps,
                )
            )(ic_set)
        return ic_set

    def get_train_ic_set(
        self,
    ) -> Float[Array, "num_train_samples num_channels *num_points"]:
        """
        Use the attributes to produce the reference training initial condition set.
        """
        return self._produce_ic_set(
            stepper=self.get_ref_stepper(),
            num_samples=self.num_train_samples,
            key=jax.random.PRNGKey(self.train_seed),
        )

    def get_train_ic_set_coarse(
        self,
    ) -> Float[Array, "num_train_samples num_channels *num_points"]:
        """
        Use the attributes to produce training initial conditions with the coarse stepper instead.
        """
        return self._produce_ic_set(
            stepper=self.get_coarse_stepper(),
            num_samples=self.num_train_samples,
            key=jax.random.PRNGKey(self.train_seed),
        )

    def get_test_ic_set(
        self,
    ) -> Float[Array, "num_test_samples num_channels *num_points"]:
        """
        Use the attributes to produce the reference testing initial condition set.
        """
        return self._produce_ic_set(
            stepper=self.get_ref_stepper(),
            num_samples=self.num_test_samples,
            key=jax.random.PRNGKey(self.test_seed),
        )

    def get_test_ic_set_coarse(
        self,
    ) -> Float[Array, "num_test_samples num_channels *num_points"]:
        """
        Use the attributes to produce testing initial conditions with the coarse stepper instead.
        """
        return self._produce_ic_set(
            stepper=self.get_coarse_stepper(),
            num_samples=self.num_test_samples,
            key=jax.random.PRNGKey(self.test_seed),
        )

    def produce_data(
        self,
        *,
        stepper: BaseStepper,
        num_samples: int,
        num_warmup_steps: int,
        temporal_horizon: int,
        key: PRNGKeyArray,
    ) -> Float[Array, "num_samples temporal_horizon+1 num_channels *num_points"]:
        """
        Default generation of data:

        1. Instantiate the intial condition distribution
        2. Generate the number of initial conditions as samples requested and
           discretize them on the grid
        3. Warmup the initial conditions if necessary
        4. Rollout these initial conditions for as many time steps as in the
           configuration

        The returned array has the shape:

        (num_train_samples, train_temporal_horizon+1, num_channels,) +
        (num_points, ) * num_spatial_dims

        the last axes are as many (num_points,) axis as there are spatial
        dimensions.
        """

        ic_distribution = self.get_ic_generator()
        ic_set = ex.build_ic_set(
            ic_distribution,
            num_points=self.num_points,
            num_samples=num_samples,
            key=key,
        )
        warmed_up_ic_set = jax.vmap(
            ex.repeat(
                stepper,
                num_warmup_steps,
            )
        )(ic_set)
        trj_set = jax.vmap(
            ex.rollout(
                stepper,
                temporal_horizon,
                include_init=True,
            )
        )(warmed_up_ic_set)

        return trj_set

    def get_train_data(
        self,
    ) -> Float[
        Array, "num_train_samples train_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Use the attributes to produce the reference training data.
        """
        return self.produce_data(
            stepper=self.get_ref_stepper(),
            num_samples=self.num_train_samples,
            num_warmup_steps=self.num_warmup_steps,
            temporal_horizon=self.train_temporal_horizon,
            key=jax.random.PRNGKey(self.train_seed),
        )

    def get_train_data_coarse(
        self,
    ) -> Float[
        Array, "num_train_samples train_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Use the attributes to produce training data with the coarse stepper instead.
        """
        return self.produce_data(
            stepper=self.get_coarse_stepper(),
            num_samples=self.num_train_samples,
            num_warmup_steps=self.num_warmup_steps,
            temporal_horizon=self.train_temporal_horizon,
            key=jax.random.PRNGKey(self.train_seed),
        )

    def get_test_data(
        self,
    ) -> Float[
        Array, "num_test_samples test_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Use the attributes to produce the reference testing data.
        """
        return self.produce_data(
            stepper=self.get_ref_stepper(),
            num_samples=self.num_test_samples,
            num_warmup_steps=self.num_warmup_steps,
            temporal_horizon=self.test_temporal_horizon,
            key=jax.random.PRNGKey(self.test_seed),
        )

    def get_test_data_coarse(
        self,
    ) -> Float[
        Array, "num_test_samples test_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Use the attributes to produce testing data with the coarse stepper instead.
        """
        return self.produce_data(
            stepper=self.get_coarse_stepper(),
            num_samples=self.num_test_samples,
            num_warmup_steps=self.num_warmup_steps,
            temporal_horizon=self.test_temporal_horizon,
            key=jax.random.PRNGKey(self.test_seed),
        )

    def get_ref_sample_data(
        self,
    ) -> Float[
        Array, "num_trjs_returned test_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Return a subset of the testing data, the number of samples is defined by
        the attribute `num_trjs_returned`
        """
        test_trj_set = self.get_test_data()
        test_trj_subset = test_trj_set[: self.num_trjs_returned]
        return test_trj_subset

    def get_callback_fn(self) -> tx.callback.BaseCallback:
        """
        Parse the `callbacks` attribute to a list of callable functions.
        """
        callback_configurations = self.callbacks.split(",")

        callback_fns = []

        for callback in callback_configurations:
            callback_args = callback.split(";")
            if callback_args[0] == "net":
                every = int(callback_args[1])
                callback_fns.append(tx.callback.GetNetwork(every=every, name="net"))
            elif callback_args[0] == "weight_norm":
                every = int(callback_args[1])
                callback_fns.append(
                    tx.callback.WeightNorm(every=every, name="weight_norm")
                )
            elif callback_args[0] == "metrics":
                every = int(callback_args[1])

                def metrics_callback_fn(update_i, model, data):
                    if update_i % every == 0:
                        metrics = self.perform_tests(model, remove_singleton_axis=True)
                        return {"metrics": metrics}
                    else:
                        return {"metrics": None}

                callback_fns.append(metrics_callback_fn)
            elif callback_args[0] == "sample_rollout":
                every = int(callback_args[1])

                def sample_rollout_callback_fn(update_i, model, data):
                    if update_i % every == 0:
                        sample_rollout = self.sample_trjs(model)
                        return {"sample_rollout": sample_rollout}
                    else:
                        return {"sample_rollout": None}

                callback_fns.append(sample_rollout_callback_fn)
            elif callback_args[0] == "":
                continue
            else:
                raise ValueError(f"Unknown callback '{callback}'")

        callback_fn = tx.callback.CompositeCallback(callback_fns)

        return callback_fn

    def get_trainer(self, *, train_config: str) -> tx.GeneralTrainer:
        """
        Expects a str of the defined interface for study. In the default
        configuration, it could for instance accept:

        'sup-03' for supervised rollout trainig with three rollout steps.

        Currently, the three major categories are available:

        - 'one' for one step supervised training
        - 'sup-XX' for supervised training with XX rollout steps
        - 'div-XX' for diverted chain training with XX rollout steps
        """
        train_trjs = self.get_train_data()

        # Needed for diverted chain training
        ref_stepper = self.get_ref_stepper()
        train_args = train_config.split(";")

        optimizer = self.get_optimizer()

        callback_fn = self.get_callback_fn()

        if train_args[0].lower() == "one":
            trainer = tx.trainer.SupervisedTrainer(
                train_trjs,
                optimizer=optimizer,
                num_training_steps=self.num_training_steps,
                batch_size=self.batch_size,
                num_rollout_steps=1,
                cut_bptt=False,
                time_level_weights=None,
                callback_fn=callback_fn,
            )
        elif train_args[0].lower() == "sup":
            num_rollout_steps = int(train_args[1])
            if len(train_args) > 2:
                cut_bptt = train_args[2].lower() == "true"
            else:
                cut_bptt = False
            trainer = tx.trainer.SupervisedTrainer(
                train_trjs,
                optimizer=optimizer,
                num_training_steps=self.num_training_steps,
                batch_size=self.batch_size,
                num_rollout_steps=num_rollout_steps,
                cut_bptt=cut_bptt,
                time_level_weights=None,
                callback_fn=callback_fn,
            )
        elif train_args[0].lower() == "div":
            num_rollout_steps = int(train_args[1])
            if len(train_args) > 2:
                cut_bptt = train_args[2].lower() == "true"
            else:
                cut_bptt = False
            if len(train_args) > 3:
                cut_div_chain = train_args[3].lower() == "true"
            else:
                cut_div_chain = False
            trainer = tx.trainer.DivertedChainBranchOneTrainer(
                train_trjs,
                ref_stepper=ref_stepper,
                optimizer=optimizer,
                num_training_steps=self.num_training_steps,
                batch_size=self.batch_size,
                num_rollout_steps=num_rollout_steps,
                cut_bptt=cut_bptt,
                cut_div_chain=cut_div_chain,
                time_level_weights=None,
                callback_fn=callback_fn,
            )
        else:
            raise ValueError(f"Unknown training argument '{train_config}'")

        return trainer

    def get_activation(
        self,
        activation: str,
    ) -> Callable:
        """
        Parse a string to a callable activation function.
        """
        activation_fn_name = activation.split(";")[0]
        activation_fn = activation_fn_dict[activation_fn_name.lower()](activation)
        return activation_fn

    def get_network(
        self,
        network_config: str,
        key: PRNGKeyArray,
    ) -> eqx.Module:
        """
        Parse the `network_config` to the corresponding neural architectue and
        instantiate it, use the `key` to initialize the parameters.

        "Conv;34;10;relu" for a feedforward convolutional network with 34 hidden
        channels, 10 hidden layers, and the ReLU activation function.

        Currently, the following constructors are available:

        - `Conv;HIDDEN_CHANNELS;DEPTH;ACTIVATION`: A feedforward
            convolutional network with `DEPTH` hidden layers of `WIDTH` size.
            Each layer transition except for the last uses `ACTIVATION`. The
            effective receptive field is `DEPTH + 1`
        - `Res;WIDTH;BLOCKS;ACTIVATION`: A classical/legacy ResNet with
            post-activation and no normalization scheme. Each residual block has
            two convolutions and operates at `WIDTH` channel size. The
            `ACTIVATION` follows each of the convolutions in the residual block.
            There are `BLOCKS` number of residual blocks. Lifting and projection
            are point-wise linear convolutions (=1x1 convs).
        - `UNet;WIDTH;LEVELS;ACTIVATION`: A classical UNet using double
            convolution blocks with group activation in-between (number of
            groups is set to one). `WIDTH` describes the hidden layer's size on
            the highest resolution level. `LEVELS` indicates the number of times
            the spatial resolution is halved by a factor of two while the
            channel count doubles. Skip connections exist between the encoder
            and decoder part of the network.
        - `Dil;DIL-FACTOR;WIDTH;BLOCKS;ACTIVATION`: Similar to the
            classical post-activation ResNet but uses a series of stacked
            convolutions of different dilation rates. Each convolution is
            followed by a group normalization (number of groups is set to one)
            and the `ACTIVATION`. `DIL-FACTOR` of 1 refers to one convolution of
            dilation rate 1. If it is set to 2, this refers to three
            convolutions of rates [1, 2, 1]. If it is 3, then this is [1, 2, 4,
            2, 1], etc.
        - `FNO;MODES;WIDTH;BLOCKS;ACTIVATION`: A vanilla FNO using spectral
            convolutions with `MODES` equally across all spatial dimensions.
            Each block operates at `WIDTH` channel size and has one spectral
            convolution with a point-wise linear bypass. The activation is
            applied to the sum of spectral convolution and bypass result. There
            are `BLOCKS` total blocks. Lifting and projection are point-wise
            linear (=1x1) convolutions.
        - `MLP;WIDTH;DEPTH;ACTIVATION`: A multi-layer perceptron with `DEPTH`
            hidden layers of `WIDTH` size. Each layer transition except for the
            last uses `ACTIVATION`. Channel and spatial axes are flattened into
            one feature axis. Hence, the MLP is hard-coded to one specific
            resolution.
        - `Pure;KERNEL_SIZE`: A purely linear convolution (with no bias) with
            kernel size `KERNEL_SIZE`. Use this to learn finite difference
            stencils. It has as many learnable parameters as the kernel size.
        - `MoRes;WIDTH;BLOCKS;ACTIVATION`: A modern ResNet using pre-activation
            and group normalization. Each residual block has two convolutions
            and operates at `WIDTH` channel size. The `ACTIVATION` follows each
            of the convolutions in the residual block. There are `BLOCKS` number
            of residual blocks. Lifting and projection are point-wise linear
            convolutions (=1x1 convs).
        - `MoUNet;WIDTH;LEVELS;ACTIVATION`: A modern UNet using two resnet
            blocks per level. `WIDTH` describes the hidden layer's size on the
            highest resolution level. `LEVELS` indicates the number of times the
            spatial resolution is halved by a factor of two while the channel
            count doubles. Skip connections exist between the encoder and
            decoder part of the network.

        The `key` is used to initialize the parameters of the neural network.

        To registor your custom architecture use the `arch_extensions`
        dictionary.

        Returns:

        - `network`: eqx.Module, the neural architecture
        """
        network_args = network_config.split(";")

        network_name = network_args[0]
        activation_fn_config = network_args[-1]
        activation_fn = self.get_activation(activation_fn_config)

        network = architecture_dict[network_name.lower()](
            network_config,
            self.num_spatial_dims,
            self.num_points,
            self.num_channels,
            activation_fn,
            key,
        )

        return network

    def get_neural_stepper(
        self, *, task_config: str, network_config: str, key: PRNGKeyArray
    ) -> eqx.Module:
        """
        Use the `network_config` to instantiate the neural architecture with
        `key` for the initial parameters. Then use the `task_config` to
        determine the wrapper around the neural architecture.

        If the `task_config` is 'predict', the neural architecture is returned
        directly.

        If the `task_config` is 'correct;XX', the neural architecture is wrapped
        in a `CorrectedStepper` with `XX` as the mode. Supported modes are:

        - `sequential`
        - `parallel`
        - `sequential_with_bypass`
        """
        network = self.get_network(network_config, key)

        task_args = task_config.split(";")
        if task_args[0] == "predict":
            neural_stepper = network
        elif task_args[0] == "correct":
            coarse_stepper = self.get_coarse_stepper()
            neural_stepper = CorrectedStepper(
                coarse_stepper=coarse_stepper,
                network=network,
                mode=task_args[1],
            )
        else:
            raise ValueError("Unknown task argument")

        return neural_stepper

    def get_parameter_count(
        self,
        network_config: str,
    ) -> int:
        """
        Count the number of parameters associated with `network_config` str.

        Note that this depends on `self.num_spatial_dims`, `self.num_channels,
        and in some cases (so far only the MLP) on `self.num_points`.
        """
        neural_stepper = self.get_neural_stepper(
            task_config="predict",  # Gives pure network without any arrays in the coarse stepper mistakingly considered as parameters
            network_config=network_config,
            key=jax.random.PRNGKey(0),  # Does not matter
        )
        return pdeqx.count_parameters(neural_stepper)

    def get_receptive_field(
        self,
        *,
        network_config: str,
        task_config: str,
    ) -> tuple[tuple[int, int], ...]:
        """
        Return the receptive field of the neural architecture for the given
        configuration.
        """
        neural_stepper = self.get_neural_stepper(
            task_config=task_config,
            network_config=network_config,
            key=jax.random.PRNGKey(0),  # Does not matter
        )
        return neural_stepper.receptive_field

    def load_model(
        self,
        path,
        *,
        num_seeds: int,
        task_config: str,
        network_config: str,
        remove_singleton_axis: bool = True,
    ) -> eqx.Module:
        """
        Load the model from the given path.
        """

        def get_stepper(i):
            return self.get_neural_stepper(
                task_config=task_config,
                network_config=network_config,
                key=jax.random.PRNGKey(i),  # Does not matter
            )

        if num_seeds == 1 and remove_singleton_axis:
            neural_stepper = get_stepper(0)
        else:
            neural_stepper = eqx.filter_vmap(get_stepper)(jnp.arange(num_seeds))
        neural_stepper = eqx.tree_deserialise_leaves(path, neural_stepper)
        return neural_stepper

    def full_loss(
        self,
        model: eqx.Module,
        *,
        train_config: str,
    ) -> float:
        """
        Computes the loss of the model on the entire training set in the
        configuration denoted by `train_config`.
        """
        trainer = self.get_trainer(train_config=train_config)
        loss = trainer.full_loss(model)
        return loss

    def perform_test_rollout(
        self,
        neural_stepper: eqx.Module,
        mean_error_fn: Callable = lambda pred, ref: ex.metrics.mean_metric(
            ex.metrics.nRMSE,
            pred,
            ref,
        ),
    ) -> Float[Array, "test_temporal_horizon"]:
        """
        Rollout the neural stepper starting from the test initial condition and
        compare it to the reference trajectory.
        """
        test_trjs = self.get_test_data()
        test_ics = test_trjs[:, 0]
        ref_trjs = test_trjs[:, 1:]
        pred_trjs = jax.vmap(
            ex.rollout(
                neural_stepper,
                self.test_temporal_horizon,
                include_init=False,
            )
        )(test_ics)

        error_rollout = jax.vmap(
            mean_error_fn,
            in_axes=1,  # over the temporal axis
        )(pred_trjs, ref_trjs)

        return error_rollout

    def get_metric_fns(
        self,
    ) -> dict[
        str,
        Callable[
            [
                Float[
                    Array,
                    "num_samples num_channels *num_points",
                ],
                Float[Array, "num_samples num_channels *num_points"],
            ],
            float,
        ],
    ]:
        """
        Return a dictionary with all metric functions according to the
        `report_metrics` attribute.
        """
        metric_fn_dict = {}

        for metric_config in self.report_metrics.split(","):
            metric_args = metric_config.split(";")
            metric_name = metric_args[0]
            metric_constructor = metric_dict[metric_name]
            metric_fn = metric_constructor(metric_config)
            metric_fn_dict[metric_config] = metric_fn

        return metric_fn_dict

    def perform_tests(
        self,
        neural_stepper: eqx.Module,
        *,
        remove_singleton_axis: bool = False,
    ) -> dict[str, Float[Array, "test_temporal_horizon"]]:
        """
        Computes all metrics according to the `report_metrics` attribute.
        """
        metric_function_dict = self.get_metric_fns()

        results = {}

        for metric_config, func in metric_function_dict.items():
            exec_func = lambda model: self.perform_test_rollout(model, func)
            if remove_singleton_axis:
                # add singleton axis for compatibility
                results[metric_config] = exec_func(neural_stepper)[None]
            else:
                results[metric_config] = eqx.filter_vmap(exec_func)(neural_stepper)

        return results

    def perform_tests_on_rollout(
        self,
        neural_rollout: Union[
            Float[
                Array,
                "num_samples test_temporal_horizon num_channels *num_points",
            ],
            Float[
                Array,
                "num_seeds num_samples test_temporal_horizon num_channels *num_points",
            ],
        ],
        test_data_no_init: Optional[
            Float[
                Array,
                "num_samples test_temporal_horizon num_channels *num_points",
            ]
        ] = None,
    ) -> Union[
        dict[str, Float[Array, "test_temporal_horizon"]],
        dict[str, Float[Array, "num_seeds test_temporal_horizon"]],
    ]:
        """
        Compute all error metrics of the `report_metrics` attribute on an
        externally produce rollout.

        !!! tip

            Use this function to benchmark external models by producing the
            initial states with `scenario.get_test_ic_set()`, roll them out in
            their respective framework for `test_temporal_horizon` steps, and
            then call this function on the produced rollout. (Some frameworks
            require different array formats, e.g.,
            [TensorFlow](https://github.com/tensorflow/tensorflow) and
            [Flax](https://github.com/google/flax) are typically channels-last.
            Hence, some reshaping might be necessary.)

        !!! warning

            The `neural_rollout` must **not** contain the initial conditions as
            the zeroth frame.

        **Arguments:**

        - `neural_rollout`: The neural rollout to be tested.
        - `test_data_no_init`: The test data without the initial conditions.
            If not provided, the test data is procedurally generated.

        **Returns:**

        - `results`: A dictionary with the metric names as keys and the error
            rollouts. The rollout arrays always have a leading `num_seeds` axis,
            even if the `neural_rollout` did not. This is to ensure
            compatibility with the `perform_tests` function. Certainly, this
            will be a singleton axis if `neural_rollout` did not have a leading
            axis.
        """
        if test_data_no_init is None:
            test_data_no_init = self.get_test_data()[:, 1:]

        if neural_rollout.shape == test_data_no_init.shape:
            multi_seed = False
        else:
            if neural_rollout.shape[1:] == test_data_no_init.shape:
                multi_seed = True
            else:
                raise ValueError(
                    f"""The shape of the neural rollout is {neural_rollout.shape}
                    and the shape of the test data is {test_data_no_init.shape}. They should
                    either be the same or the neural rollout should have an additional
                    leading axis for the seeds."""
                )

        metric_function_dict = self.get_metric_fns()

        results = {}

        for metric_config, func in metric_function_dict.items():
            if multi_seed:
                func_vectorized = jax.vmap(
                    jax.vmap(
                        func,
                        # Vectorize over time steps
                        in_axes=(1, 1),
                    ),
                    # Vectorize over seeds (but broadcast the reference)
                    in_axes=(0, None),
                )
            else:
                func_vectorized = jax.vmap(
                    func,
                    # Vectorize over time steps
                    in_axes=(1, 1),
                )

            metric_rollout = func_vectorized(neural_rollout, test_data_no_init)
            if not multi_seed:
                # Add singletone seed axis for compatibility with `perform_tests`
                metric_rollout = metric_rollout[None]

            results[metric_config] = metric_rollout

        return results

    def sample_trjs(
        self, neural_stepper: eqx.Module
    ) -> Float[
        Array, "num_trjs_returned test_temporal_horizon+1 num_channels *num_points"
    ]:
        """
        Use the neural_stepper to produce a sample of trajectories. The initial
        conditions are the ones from the test set.
        """
        test_trjs = self.get_test_data()
        test_ics_subset = test_trjs[: self.num_trjs_returned, 0]
        sample_trj_s = jax.vmap(
            ex.rollout(
                neural_stepper,
                self.test_temporal_horizon,
                include_init=True,
            )
        )(test_ics_subset)
        return sample_trj_s

    def run_raw(
        self,
        *,
        task_config: str = "predict",
        network_config: str = "Conv;26;10;relu",
        train_config: str = "one",
        start_seed: int = 0,
        num_seeds: int = 1,
        remove_singleton_axis: bool = False,
    ):
        """
        For more details see the __call__ method.

        Use this function if you intend to wrap your run in further vmaps.

        **Returns:**

        - `trained_neural_stepper_s`: eqx.Module, the trained neural stepper
            for the scenario. If `num_seeds` is 1, the singleton dimension along
            the batch axis is removed (if `remove_singleton_axis` is True).
        - `loss_history_s`: Array, the loss history of the training. The
            shape is `(num_seeds, num_training_steps//record_loss_every)`
        - `aux_history_s`: Array, the auxiliary history of the training. The
            shape is `(num_seeds, num_training_steps)`
        - `metric_trj_s`: dict, the metrics computed on the test set. The
            keys are the metric names and the values are arrays with the shape
            `(num_seeds, test_temporal_horizon)`
        - `sample_rollout_s`: Array, the sample rollouts produced by the
            trained neural stepper. The shape is `(num_seeds, num_trjs_returned,
            test_temporal_horizon+1, num_channels, *num_points)`
        - `seeds`: Array, the seeds used for the run
        """
        trainer = self.get_trainer(train_config=train_config)

        def produce_result_one_seed(seed):
            key = jax.random.PRNGKey(seed)
            init_key, shuffle_key = jax.random.split(key, 2)
            neural_stepper = self.get_neural_stepper(
                task_config=task_config,
                network_config=network_config,
                key=init_key,
            )
            trained_neural_stepper, loss_history, aux_history = trainer(
                neural_stepper,
                shuffle_key,
                record_loss_every=self.record_loss_every,
            )

            sample_rollout = self.sample_trjs(trained_neural_stepper)

            return (
                trained_neural_stepper,
                loss_history,
                aux_history,
                # mean_nRMSE_rollout,
                sample_rollout,
            )

        seeds = start_seed + jnp.arange(num_seeds)

        # Adds additional batch axis to the output of produce_result_one_seed
        (
            trained_neural_stepper_s,
            loss_history_s,
            aux_history_s,
            # error_trj_s,
            sample_rollout_s,
        ) = eqx.filter_vmap(produce_result_one_seed)(seeds)

        metric_trj_s = self.perform_tests(trained_neural_stepper_s)

        results = (
            trained_neural_stepper_s,
            loss_history_s,
            aux_history_s,
            metric_trj_s,
            sample_rollout_s,
            seeds,
        )

        # If only one seed is considered, remove the singleton axis if requested
        if num_seeds == 1 and remove_singleton_axis:
            results = pdeqx.extract_from_ensemble(results, 0)

        return results

    def __call__(
        self,
        *,
        task_config: str = "predict",
        network_config: str = "Conv;34;10;relu",
        train_config: str = "one",
        start_seed: int = 0,
        num_seeds: int = 1,
        remove_singleton_axis: bool = True,
    ) -> tuple[pd.DataFrame, eqx.Module]:
        """
        Execute the scenario with the given attribute configuration and the
        additional configuration strings.

        **Arguments:**

        - `task_config`: What the trained neural predictor should
            represent. Can be either 'predict' or 'correct;XX' where XX is the
            mode of correction. `predict` refers to a pure neural architecture.
            The neural network will fully replace the numerical timestepper. In
            the case of `correct;XX`, the neural network interacts with a coarse
            stepper. To inference such a system after training, the
            corresponding coarse solver is needed, but is already baked into the
            returning module. Default is 'predict'.
        - `network_config`: The configuration of the neural network.
            This begins with a general architecture type, followed by a
            architecture-dependent length list of parameters. See the method
            `get_network` for the available architectures and their
            configuration. Default is 'Conv;34;10;relu' which is a feed-forward
            convolutional network with 34 hidden channels over 10 hidden layers
            and the ReLU activation function (about 30k parameters for 1D
            problems)
        - `train_config`: The training configuration. This determines
            how neural stepper and reference numerical stepper interact during
            training. See the method `get_trainer` for the available training
            configurations. Default is 'one' which refers to a one-step
            supervised approach in which one batch of samples with a length 2
            window is sampled across all initial conditions and temporal
            horizon.
        - `start_seed`: The starting seed for the random number
            generator of network initialization. Default is 0.
        - `num_seeds`: The number of seeds to use. Default is 1.
        - `remove_singleton_axis`: bool, if True and `num_seeds` is 1, the
            singleton axis resulting from the seed parallel runs is removed
            which allows to directly use the returned neural stepper. Otherwise,
            it must be wrapped in a `eqx.filter_vmap(...)`

        **Returns:**

        - `result_df`: A dataframe with the results of the
            scenario. Each row represents one seed. It contains the following
            columns:
            - 'scenario': str, the name of the scenario, created by the
                method `get_scenario_name`
            - 'task': str, the task configuration (as given in the
                argument)
            - 'train': str, the training configuration (as given in the
                argument)
            - 'net': str, the network configuration (as given in the
                argument)
            - 'seed': int, the seed used for the run (this varies
                between the rows if multiple seeds are used at the same time)
            - 'mean_nRMSE_XXXX': float, the mean nRMSE metric produced
                in an error rollout **after the training has finished**. Each
                temporal entry (staring at 1 all the way to
                `self.test_temporal_horizon`) is represented by a separate
                column.
            - `METRICS_XXXX`: float, additional metrics (e.g., mean
                correlation rollout)
            - 'train_loss_XXXXXX': float, the training loss at each
                training step. Each step is represented by a separate column
                (starting at 0 all the way to `self.num_training_steps - 1`)
            - 'aux_XXXXXX': list, the history of auxiliary information
                produced by callbacks. If there is no callback active, each
                entry is an empty dictionary.
            - 'sample_rollout_XXX': list, a list of lists representing
                the sample rollouts produced by the trained neural stepper. The
                outer list represents the different initial conditions, the
                inner lists represent the different time steps. The length of
                the outer list is given by the attribute `num_trjs_returned`. We
                use list to store (jax.)numpy arrays.
        - `trained_neural_stepper_s`: eqx.Module, the trained neural stepper
            for the scenario. This follows an structure of arrays approach to
            represent the collection of networks trained based on different
            initialization seeds. If `num_seeds` is 1 (it is only intended to
            train one network), use the `remove_singleton_axis` argument to
            remove the singleton dimension (True by default).

        !!! note
            A typical workflow is to use the functions
            `apebench.utils.melt_loss`, `apebench.utils.melt_metrics`, and
            `apebench.utils.melt_sample_rollouts` to melt the returned dataframe
            into a long format that can be used for plotting with seaborn.
        """
        (
            trained_neural_stepper_s,
            loss_history_s,
            aux_history_s,
            metric_trj_s,
            sample_rollout_s,
            seeds,
        ) = self.run_raw(
            task_config=task_config,
            network_config=network_config,
            train_config=train_config,
            start_seed=start_seed,
            num_seeds=num_seeds,
            remove_singleton_axis=False,
        )

        n_training_steps = loss_history_s.shape[-1]
        n_sample_rollouts_returned = sample_rollout_s.shape[1]

        scenario_name = self.get_scenario_name()

        metric_dicts = []
        for metric, metric_trj in metric_trj_s.items():
            metric_dicts.append(
                {
                    f"{metric}_{i+1:04d}": metric_trj[:, i]  # noqa: E226
                    for i in range(self.test_temporal_horizon)
                }
            )

        aux_dicts = []
        for i, entry in enumerate(aux_history_s):
            aux_dicts.append(
                {
                    f"aux_{i:06d}": [
                        pdeqx.extract_from_ensemble(entry, j) for j in range(num_seeds)
                    ]
                }
            )

        result_df = pd.DataFrame(
            dict(
                **{
                    "scenario": scenario_name,
                    "task": task_config,
                    "train": train_config,
                    "net": network_config,
                    "seed": seeds,
                    # Needed for being compliant with multi-experiment interface
                    "scenario_kwargs": "{}",
                },
                **{
                    key: value
                    for sub_dict in metric_dicts
                    for key, value in sub_dict.items()
                },
                **{
                    f"train_loss_{(i * self.record_loss_every):06d}": loss_history_s[
                        :, i
                    ]
                    for i in range(n_training_steps)
                },
                **{
                    key: value
                    for sub_dict in aux_dicts
                    for key, value in sub_dict.items()
                },
                **{
                    f"sample_rollout_{i:03d}": sample_rollout_s[:, i].tolist()
                    for i in range(n_sample_rollouts_returned)
                },
            )
        )

        # If there is only one seed considered, remove the singleton dimension
        # in the weight arrays
        if num_seeds == 1 and remove_singleton_axis:
            trained_neural_stepper_s = pdeqx.extract_from_ensemble(
                trained_neural_stepper_s,
                0,
            )

        return result_df, trained_neural_stepper_s



from typing import Callable, Dict

import exponax as ex

ic_dict: Dict[
    str,  # ic_name
    Callable[
        [
            str,  # ic_config
            int,  # num_spatial_dims
        ],
        ex.ic.BaseRandomICGenerator,
    ],
] = {
    "fourier": lambda ic_config, num_spatial_dims: ex.ic.RandomTruncatedFourierSeries(
        num_spatial_dims=num_spatial_dims,
        cutoff=int(ic_config.split(";")[1]),
        offset_range=(0.0, 0.0)
        if ic_config.split(";")[2].lower() == "true"
        else (-0.5, 0.5),
        max_one=ic_config.split(";")[3].lower() == "true",
    ),
    "diffused": lambda ic_config, num_spatial_dims: ex.ic.DiffusedNoise(
        num_spatial_dims=num_spatial_dims,
        intensity=float(ic_config.split(";")[1]),
        zero_mean=ic_config.split(";")[2].lower() == "true",
        max_one=ic_config.split(";")[3].lower() == "true",
    ),
    "grf": lambda ic_config, num_spatial_dims: ex.ic.GaussianRandomField(
        num_spatial_dims=num_spatial_dims,
        powerlaw_exponent=float(ic_config.split(";")[1]),
        zero_mean=ic_config.split(";")[2].lower() == "true",
        max_one=ic_config.split(";")[3].lower() == "true",
    ),
}



import argparse
import datetime
import json
import os
import shutil
import sys
from pathlib import Path

from ._run import melt_concat_from_list, run_study


def app():
    parser = argparse.ArgumentParser(
        description="Run a study by pointing to its config list"
    )
    parser.add_argument("study", type=str, help="The file containig the config list")
    parser.add_argument(
        "--gpu", type=int, help="The GPU to use for the experiment", default=0
    )
    parser.add_argument(
        "--start_seed",
        type=int,
        help="The seed to start the experiment from (helpful for seed-statistics across multiple GPUs)",
        default=None,
    )
    parser.add_argument(
        "--dont_melt_metrics",
        action="store_true",
        help="Do not melt the metrics across the experiments in the study",
        default=False,
    )
    parser.add_argument(
        "--dont_melt_loss",
        action="store_true",
        help="Do not melt the train loss across the experiments in the study",
        default=False,
    )
    parser.add_argument(
        "--melt_sample_rollouts",
        action="store_true",
        help="Melt the sample rollouts across the experiments in the study",
        default=False,
    )
    parser.add_argument(
        "--metric_name",
        type=str,
        help="The name(s) of the metric(s) to melt. For multiple use a comma separated list, e.g. 'mean_nRMSE,mean_correlation'",
        default="mean_nRMSE",
    )
    args = parser.parse_args()

    # Set the GPU to use
    os.environ["CUDA_VISIBLE_DEVICES"] = str(args.gpu)

    # TODO: change hard coded version
    apebench_version = "0.0.1"

    LOG_DIR = "logs/"
    os.makedirs(LOG_DIR, exist_ok=True)

    LOG_SUBDIR = LOG_DIR + str(datetime.datetime.now()) + "/"

    os.makedirs(LOG_SUBDIR, exist_ok=True)

    with open(LOG_SUBDIR + "apebench_version.txt", "w") as f:
        f.write(apebench_version)

    # # Save the environment file
    # environment_file = subprocess.run(
    #     ["conda", "env", "export"], capture_output=True, text=True
    # ).stdout
    # with open(LOG_SUBDIR + "environment.yml", "w") as f:
    #     f.write(environment_file)

    study_path = Path(args.study)

    if not os.path.exists(study_path):
        print(f"Error: The file '{study_path}' does not exist.")
        sys.exit(1)

    # Copy the config file to the log directory
    shutil.copy(study_path, LOG_SUBDIR)

    # Load the variable CONFIGS from the python file in experiment_path
    study_dir = os.path.dirname(study_path)
    sys.path.append(study_dir)
    study = os.path.basename(study_path)
    study_name = study.split(".")[0]
    module = __import__(study_name)
    CONFIGS = module.CONFIGS

    if args.start_seed is not None:
        print("Overwriting all start seeds to", args.start_seed)
        new_configs = []
        for config in CONFIGS:
            config["start_seed"] = args.start_seed
            new_configs.append(config)
        CONFIGS = new_configs

    BASE_PATH = "results/"

    os.makedirs(BASE_PATH, exist_ok=True)

    raw_file_list, network_weights_list = run_study(
        CONFIGS,
        BASE_PATH,
    )

    MELTED_DIR = "melted/"

    os.makedirs(MELTED_DIR, exist_ok=True)

    MELTED_SUBDIR = MELTED_DIR + study_name + "/"

    os.makedirs(MELTED_SUBDIR, exist_ok=True)

    with open(MELTED_SUBDIR + "network_weights_list.json", "w") as f:
        dump = json.dumps([str(r) for r in network_weights_list])
        f.write(dump)

    with open(MELTED_SUBDIR + "raw_file_list.json", "w") as f:
        dump = json.dumps([str(r) for r in raw_file_list])
        f.write(dump)

    melt_concat_from_list(
        raw_file_list,
        MELTED_SUBDIR,
        metric_name=args.metric_name.split(","),
        do_metrics=not args.dont_melt_metrics,
        do_loss=not args.dont_melt_loss,
        do_sample_rollouts=args.melt_sample_rollouts,
    )



import exponax as ex

from ..._base_scenario import BaseScenario


class Linear(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.1

    a_coefs: tuple[float, ...] = (0.0, -0.25, 0.0, 0.0, 0.0)
    coarse_proportion: float = 0.5

    def get_ref_stepper(self):
        return ex.stepper.generic.GeneralLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt,
            linear_coefficients=self.a_coefs,
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.generic.GeneralLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=self.dt * self.coarse_proportion,
            linear_coefficients=self.a_coefs,
        )

    def get_scenario_name(self) -> str:
        active_indices = []
        for i, a in enumerate(self.a_coefs):
            if a != 0.0:
                active_indices.append(i)
        return f"{self.num_spatial_dims}d_phy_lin_{'_'.join(str(i) for i in active_indices)}"


class LinearSimple(Linear):
    linear_coef: float = -0.25
    linear_term_order: int = 1

    def __post_init__(self):
        self.a_coefs = (0.0,) * self.linear_term_order + (self.linear_coef,)


class Advection(Linear):
    advection_coef: float = -0.25

    def __post_init__(self):
        self.a_coefs = (0.0, self.advection_coef, 0.0, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_adv"


class Diffusion(Linear):
    diffusion_coef: float = 0.008

    def __post_init__(self):
        self.a_coefs = (0.0, 0.0, self.diffusion_coef, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_diff"


class AdvectionDiffusion(Linear):
    advection_coef: float = -0.25
    diffusion_coef: float = 0.008

    def __post_init__(self):
        self.a_coefs = (0.0, self.advection_coef, self.diffusion_coef, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_adv_diff"


class Dispersion(Linear):
    dipersion_coef: float = 0.00025
    dt: float = 0.001  # Overwrite

    def __post_init__(self):
        self.a_coefs = (0.0, 0.0, 0.0, self.dipersion_coef, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_disp"


class HyperDiffusion(Linear):
    hyper_diffusion_coef: float = -0.000075
    dt: float = 0.00001  # Overwrite

    def __post_init__(self):
        self.a_coefs = (0.0, 0.0, 0.0, 0.0, self.hyper_diffusion_coef)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_hyp_diff"


class FirstFour(Linear):
    advection_coef: float = -2500.0
    diffusion_coef: float = 80.0
    dispersion_coef: float = 0.025
    hyp_diffusion_coef: float = -0.000075
    dt: float = 0.00001  # Overwrite

    def __post_init__(self):
        self.a_coefs = (
            0.0,
            self.advection_coef,
            self.diffusion_coef,
            self.dispersion_coef,
            self.hyp_diffusion_coef,
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_four"



from ._activation_function import activation_fn_dict
from ._architectures import architecture_dict
from ._initial_condition import ic_dict
from ._metrics import metric_dict
from ._optimization import lr_scheduler_dict, optimizer_dict

__all__ = [
    "metric_dict",
    "lr_scheduler_dict",
    "optimizer_dict",
    "ic_dict",
    "activation_fn_dict",
    "architecture_dict",
]



import exponax as ex

from ..._base_scenario import BaseScenario


class Linear(BaseScenario):
    gammas: tuple[float, ...] = (0.0, -4.0, 0.0, 0.0, 0.0)
    coarse_proportion: float = 0.5

    def get_ref_stepper(self):
        return ex.stepper.generic.DifficultyLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            linear_difficulties=self.gammas,
        )

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return ex.stepper.generic.DifficultyLinearStepper(
            num_spatial_dims=self.num_spatial_dims,
            num_points=self.num_points,
            linear_difficulties=tuple(f * self.coarse_proportion for f in self.gammas),
        )

    def get_scenario_name(self) -> str:
        active_indices = []
        for i, a in enumerate(self.gammas):
            if a != 0.0:
                active_indices.append(i)
        return f"{self.num_spatial_dims}d_diff_linear_{'_'.join(str(i) for i in active_indices)}"


class LinearSimple(Linear):
    linear_gamma: float = -4.0
    linear_term_order: int = 1

    def __post_init__(self):
        self.gammas = (0.0,) * self.linear_term_order + (self.linear_gamma,)


class Advection(Linear):
    advection_gamma: float = -4.0

    def __post_init__(self):
        self.gammas = (0.0, self.advection_gamma, 0.0, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_adv"


class Diffusion(Linear):
    diffusion_gamma: float = 4.0

    def __post_init__(self):
        self.gammas = (0.0, 0.0, self.diffusion_gamma, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_diff"


class AdvectionDiffusion(Linear):
    advection_gamma: float = -4.0
    diffusion_gamma: float = 4.0

    def __post_init__(self):
        self.gammas = (0.0, self.advection_gamma, self.diffusion_gamma, 0.0, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_adv_diff"


class Dispersion(Linear):
    dispersion_gamma: float = 4.0

    def __post_init__(self):
        self.gammas = (0.0, 0.0, 0.0, self.dispersion_gamma, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_disp"


class HyperDiffusion(Linear):
    hyp_diffusion_gamma: float = -4.0

    def __post_init__(self):
        self.gammas = (0.0, 0.0, 0.0, 0.0, self.hyp_diffusion_gamma, 0.0)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_hyp_diff"


class FirstFour(Linear):
    advection_gamma: float = -4.0
    diffusion_gamma: float = 4.0
    dispersion_gamma: float = 4.0
    hyp_diffusion_gamma: float = -4.0

    def __post_init__(self):
        self.gammas = (
            0.0,
            self.advection_gamma,
            self.diffusion_gamma,
            self.dispersion_gamma,
            self.hyp_diffusion_gamma,
        )

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_four"



import equinox as eqx
import exponax
import jax
import jax.numpy as jnp

import apebench

# import pytest


def test_external_metric_computation_single_seed():
    adv_scenario = apebench.scenarios.difficulty.Advection()

    # Just use the randomly initialized model
    neural_stepper = adv_scenario.get_neural_stepper(
        task_config="predict",
        network_config="mlp;16;4;relu",
        key=jax.random.PRNGKey(0),
    )

    # Manual path: (1) get test ic, (2) rollout model, (3) compute metrics
    test_ic = adv_scenario.get_test_ic_set()
    neural_rollout = jax.vmap(
        exponax.rollout(neural_stepper, adv_scenario.test_temporal_horizon)
    )(test_ic)
    manual_metrics = adv_scenario.perform_tests_on_rollout(neural_rollout)

    # Automated path: Directly use `perform_tests`
    automated_metrics = adv_scenario.perform_tests(
        neural_stepper,
        remove_singleton_axis=True,
    )

    manual_nRMSE = manual_metrics["mean_nRMSE"]
    automated_nRMSE = automated_metrics["mean_nRMSE"]

    assert jnp.array_equal(manual_nRMSE, automated_nRMSE)
    # assert manual_nRMSE == pytest.approx(automated_nRMSE, rel=1e-6, abs=1e-12)


def test_external_metric_computation_multiple_seeds():
    adv_scenario = apebench.scenarios.difficulty.Advection()

    # Just use the randomly initialized model, but multiple seeds
    neural_stepper_s = eqx.filter_vmap(
        lambda s: adv_scenario.get_neural_stepper(
            task_config="predict",
            network_config="mlp;16;4;relu",
            key=jax.random.PRNGKey(s),
        )
    )(jnp.arange(3))

    # Manual path: (1) get test ic, (2) rollout model, (3) compute metrics
    test_ic = adv_scenario.get_test_ic_set()
    # neural_rollout = jax.vmap(
    #     exponax.rollout(neural_stepper, adv_scenario.test_temporal_horizon)
    # )(test_ic)
    neural_rollout_s = eqx.filter_vmap(
        lambda ns: jax.vmap(exponax.rollout(ns, adv_scenario.test_temporal_horizon))(
            test_ic
        )
    )(neural_stepper_s)

    manual_metrics = adv_scenario.perform_tests_on_rollout(neural_rollout_s)

    # Automated path: Directly use `perform_tests`
    automated_metrics = adv_scenario.perform_tests(
        neural_stepper_s,
        remove_singleton_axis=False,
    )

    manual_nRMSE = manual_metrics["mean_nRMSE"]
    automated_nRMSE = automated_metrics["mean_nRMSE"]

    assert jnp.array_equal(manual_nRMSE, automated_nRMSE)



"""
Scenarios using the general convection interface (growing channels over spatial
dimensions)
"""
import exponax as ex

from ..._base_scenario import BaseScenario


class Convection(BaseScenario):
    gammas: tuple[float, ...] = (0.0, 0.0, 1.5, 0.0, 0.0)
    convection_delta: float = -1.5
    conservative: bool = True

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: float = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        self.num_channels = self.num_spatial_dims  # Overwrite

    def _build_stepper(self, gammas, delta):
        substepped_gammas = tuple(g / self.num_substeps for g in gammas)
        substepped_delta = delta / self.num_substeps

        substepped_stepper = ex.stepper.generic.DifficultyConvectionStepper(
            self.num_spatial_dims,
            self.num_points,
            linear_difficulties=substepped_gammas,
            # Need minus to move the convection to the right hand side
            convection_difficulty=-substepped_delta,
            conservative=self.conservative,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.gammas, self.convection_delta)

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(
            tuple(f * self.coarse_proportion for f in self.gammas),
            self.coarse_proportion * self.convection_delta,
        )

    def get_scenario_name(self) -> str:
        active_indices = []
        for i, a in enumerate(self.gammas):
            if a != 0.0:
                active_indices.append(i)
        return f"{self.num_spatial_dims}d_diff_conv_{'_'.join(str(i) for i in active_indices)}"


class Burgers(Convection):
    convection_delta: float = -1.5  # Overwrite
    diffusion_gamma: float = 1.5

    def __post_init__(self):
        self.gammas = (0.0, 0.0, self.diffusion_gamma, 0.0, 0.0)
        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_burgers"


class KuramotoSivashinskyConservative(Convection):
    convection_delta: float = -1.0  # Overwrite
    diffusion_gamma: float = -2.0  # Negative diffusion; producing energy
    hyp_diffusion_gamma: float = -18.0

    num_warmup_steps: int = 500  # Overwrite
    vlim: tuple[float, float] = (-2.5, 2.5)  # Overwrite

    report_metrics: str = "mean_nRMSE,mean_correlation"  # Overwrite

    def __post_init__(self):
        if self.num_spatial_dims != 1:
            raise ValueError("KuramotoSivashinskyConservative is only defined for 1D")
        self.gammas = (0.0, 0.0, self.diffusion_gamma, 0.0, self.hyp_diffusion_gamma)
        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_diff_ks_cons"



from math import pi

import exponax as ex

from ..._base_scenario import BaseScenario


class DecayingTurbulence(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.1

    diffusivity: float = 1e-4

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: int = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        if self.num_spatial_dims != 2:
            raise ValueError(
                "Decaying turbulence is only defined for 2 spatial dimensions"
            )

    def _build_stepper(self, dt):
        substepped_stepper = ex.stepper.NavierStokesVorticity(
            self.num_spatial_dims,
            self.domain_extent,
            self.num_points,
            dt / self.num_substeps,
            diffusivity=self.diffusivity,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_decay_turb"


class KolmogorovFlow(BaseScenario):
    domain_extent: float = 2 * pi
    dt: float = 0.1

    diffusivity: float = 1e-2  # Just Re=100 to have it almost scale-resolved

    injection_mode: int = 4
    injection_scale: float = 1.0
    drag: float = -0.1

    num_substeps: int = 20

    coarse_proportion: float = 0.5

    num_warmup_steps: int = 500  # Overwrite

    vlim: tuple[float, float] = (-10.0, 10.0)  # Overwrite

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: int = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        if self.num_spatial_dims != 2:
            raise ValueError("Kolmogorov Flow is only defined for 2 spatial dimensions")

    def _build_stepper(self, dt):
        substepped_stepper = ex.stepper.KolmogorovFlowVorticity(
            self.num_spatial_dims,
            self.domain_extent,
            self.num_points,
            dt / self.num_substeps,
            diffusivity=self.diffusivity,
            injection_mode=self.injection_mode,
            injection_scale=self.injection_scale,
            drag=self.drag,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_kolm_flow"



from typing import Callable, Dict

import jax

activation_fn_dict: Dict[
    str,
    Callable[
        [
            str,  # activation_function_config
        ],
        Callable,  # activation_fn
    ],
] = {
    "relu": lambda activation_fn_config: jax.nn.relu,
    "sigmoid": lambda activation_fn_config: jax.nn.sigmoid,
    "tanh": lambda activation_fn_config: jax.nn.tanh,
    "swish": lambda activation_fn_config: jax.nn.swish,
    "gelu": lambda activation_fn_config: jax.nn.gelu,
    "identity": lambda activation_fn_config: lambda x: x,
}



import pytest

import apebench


def test_simple():
    apebench.scenarios.difficulty.Advection()


@pytest.mark.parametrize(
    "name",
    list(apebench.scenarios.scenario_dict.keys()),
)
def test_builtin_scenarios(name: str):
    # Some scenarios might not work in 1d, (which is the default number of spatial dims)
    try:
        scene = apebench.scenarios.scenario_dict[name]()
    except ValueError:
        return

    ref = scene.get_ref_sample_data()

    del ref


@pytest.mark.parametrize(
    "name,num_spatial_dims",
    [
        (name, num_spatial_dims)
        for name in [
            "phy_adv",
            "norm_adv",
            "diff_adv",
        ]
        for num_spatial_dims in [1, 2, 3]
    ],
)
def test_simple_training(name: str, num_spatial_dims: int):
    NUM_TRAIN_SAMPLES = 5
    NUM_TEST_SAMPLES = 5
    NUM_POINTS = 15
    OPTIM_CONFIG = "adam;10;constant;1e-4"

    # Some scenarios might not work in 1d, (which is the default number of spatial dims)
    try:
        scene = apebench.scenarios.scenario_dict[name](
            num_spatial_dims=num_spatial_dims,
            num_train_samples=NUM_TRAIN_SAMPLES,
            num_test_samples=NUM_TEST_SAMPLES,
            num_points=NUM_POINTS,
            optim_config=OPTIM_CONFIG,
        )
    except ValueError:
        return

    NETWORK_CONFIG = "Conv;10;2;relu"

    data, trained_net = scene(
        network_config=NETWORK_CONFIG,
    )

    del data, trained_net



from typing import Union

import jax
import jax.numpy as jnp
import pandas as pd
from scipy.stats import gmean

from ._base_scenario import BaseScenario

BASE_NAMES = [
    "seed",
    "scenario",
    "task",
    "net",
    "train",
    "scenario_kwargs",
]
BASE_NAMES_NO_TRAIN = [
    "seed",
    "scenario",
    "task",
    "net",
    "scenario_kwargs",
]


def melt_data(
    wide_data: pd.DataFrame,
    quantity_name: Union[str, list[str]],
    uniquifier_name: str,
    *,
    base_columns: list[str] = BASE_NAMES,
) -> pd.DataFrame:
    """
    Melt a wide APEBench result DataFrame into a long format suitable for
    visualization (e.g. with seaborn or plotly).

    **Arguments:**

    * `wide_data`: The wide DataFrame to melt, must contain `quantity_name` and
        `base_columns` as columns.
    * `quantity_name`: The name of the column(s) to melt.
    * `uniquifier_name`: The name of the column that will be used to uniquely
        identify the melted rows.
    * `base_columns`: The columns that should be kept as is in the melted
        DataFrame.

    **Returns:**

    * A long DataFrame with the same columns as `base_columns` and the melted
        `quantity_name`.
    """
    if isinstance(quantity_name, str):
        quantity_name = [
            quantity_name,
        ]
    data_melted = pd.wide_to_long(
        wide_data,
        stubnames=quantity_name,
        i=base_columns,
        j=uniquifier_name,
        sep="_",
    )
    data_melted = data_melted[quantity_name]
    data_melted = data_melted.reset_index()

    return data_melted


def melt_metrics(
    wide_data: pd.DataFrame,
    metric_name: Union[str, list[str]] = "mean_nRMSE",
) -> pd.DataFrame:
    """
    Melt the metrics from a wide DataFrame.
    """
    return melt_data(
        wide_data,
        quantity_name=metric_name,
        uniquifier_name="time_step",
    )


def melt_loss(wide_data: pd.DataFrame, loss_name: str = "train_loss") -> pd.DataFrame:
    """
    Melt the loss from a wide DataFrame.
    """
    return melt_data(
        wide_data,
        quantity_name=loss_name,
        uniquifier_name="update_step",
    )


def melt_sample_rollouts(
    wide_data: pd.DataFrame,
    sample_rollout_name: str = "sample_rollout",
) -> pd.DataFrame:
    """
    Melt the sample rollouts from a wide DataFrame.
    """
    return melt_data(
        wide_data,
        quantity_name=sample_rollout_name,
        uniquifier_name="sample_index",
    )


def split_train(
    metric_data: pd.DataFrame,
) -> pd.DataFrame:
    """
    Decode the `train` column into `category`, `type`, and `rollout` columns.
    """
    metric_data["category"] = metric_data["train"].apply(lambda x: x.split(";")[0])
    metric_data["type"] = metric_data["category"].apply(
        lambda x: "sup" if x in ["one", "sup"] else "div"
    )
    metric_data["rollout"] = metric_data["train"].apply(
        lambda x: int(x.split(";")[1]) if x != "one" else 1
    )

    return metric_data


def aggregate_gmean(
    metric_data: pd.DataFrame,
    *,
    up_to: int = 100,
    grouping_cols: list[str] = BASE_NAMES,
) -> pd.DataFrame:
    """
    Aggregate an error rollout over time via the geometric mean.

    Args:

    * `metric_data`: The DataFrame to aggregate, must contain `grouping_cols`
        and `mean_nRMSE` as columns. When grouped by `grouping_cols`, the groups
        shall only contain values at different time steps.
    * `up_to`: The time step up to which to aggregate. (inclusive)
    * `grouping_cols`: The columns to group by.

    Returns:

    * A DataFrame with the new column `gmean_mean_nRMSE` containing the
        geometric mean of the `mean_nRMSE` values up to `up_to` for each group.
    """
    return (
        metric_data.query(f"time_step <= {up_to}")
        .groupby(grouping_cols)
        .agg(gmean_mean_nRMSE=("mean_nRMSE", gmean))
        .reset_index()
    )


def relative_by_config(
    data: pd.DataFrame,
    *,
    grouping_cols: list[str] = BASE_NAMES_NO_TRAIN,
    norm_query: str = "train == 'one'",
    value_col: str = "mean_nRMSE",
    suffix: str = "_rel",
) -> pd.DataFrame:
    def relativate_fn(sub_df):
        rel = sub_df.query(norm_query)[value_col]
        if len(rel) != 1:
            raise ValueError(
                f"Expected exactly one row to match {norm_query}, got {len(rel)}"
            )
        col = sub_df[value_col] / rel.values[0]
        sub_df[f"{value_col}{suffix}"] = col
        return sub_df

    return data.groupby(grouping_cols).apply(relativate_fn).reset_index(drop=True)


def read_in_kwargs(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse the `scenario_kwargs` column of a DataFrame and add the parsed entries
    as new columns.

    Requires that the dictionary in `scenario_kwargs` has the same keys for all
    rows.
    """
    col = df["scenario_kwargs"].apply(eval)
    entries = list(col[0].keys())
    for entry in entries:
        df[entry] = col.apply(lambda x: x[entry])
    return df


def count_nan_trjs(trjs: jax.Array) -> int:
    """
    Computes the number of trajectories that contain at least one NaN value.
    """

    def has_nan(trj):
        if jnp.sum(jnp.isnan(trj)) > 0:
            return 1
        else:
            return 0

    mask = [has_nan(trj) for trj in trjs]

    return sum(mask)


def check_for_nan(scene: BaseScenario):
    """
    Check for NaNs in the train and test data of a scenario. Also checks the
    train and test data set produced by the coarse stepper if the scenario
    supports a correction mode. Raises an AssertionError if NaNs are found.
    """
    train_data = scene.get_train_data()

    train_num_nans = count_nan_trjs(train_data)
    assert (
        train_num_nans == 0
    ), f"Train data has {train_num_nans} trajectories with NaNs"

    del train_data

    test_data = scene.get_test_data()

    test_num_nans = count_nan_trjs(test_data)
    assert test_num_nans == 0, f"Test data has {test_num_nans} trajectories with NaNs"

    del test_data

    try:
        # Some scenarios might not support a correction mode
        train_data_coarse = scene.get_train_data_coarse()

        train_num_nans_coarse = count_nan_trjs(train_data_coarse)
        assert (
            train_num_nans_coarse == 0
        ), f"Train data coarse has {train_num_nans_coarse} trajectories with NaNs"

        del train_data_coarse

        test_data_coarse = scene.get_test_data_coarse()

        test_num_nans_coarse = count_nan_trjs(test_data_coarse)
        assert (
            test_num_nans_coarse == 0
        ), f"Test data coarse has {test_num_nans_coarse} trajectories with NaNs"

        del test_data_coarse
    except NotImplementedError:
        return



from typing import Callable, Dict

import exponax as ex
from jaxtyping import Array, Float

metric_dict: Dict[
    str,
    Callable[
        [
            str,  # metric_config
        ],
        Callable[
            [
                Float[Array, "batch channel ... N"],  # batched pred
                Float[Array, "batch channel ... N"],  # batched target
            ],
            float,
        ],
    ],
] = {
    "mean_MAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.MAE,
        pred,
        ref,
    ),
    "mean_nMAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.nMAE,
        pred,
        ref,
    ),
    "mean_sMAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.sMAE,
        pred,
        ref,
    ),
    "mean_MSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.MSE,
        pred,
        ref,
    ),
    "mean_nMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.nMSE,
        pred,
        ref,
    ),
    "mean_sMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.sMSE,
        pred,
        ref,
    ),
    "mean_RMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.RMSE,
        pred,
        ref,
    ),
    "mean_nRMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.nRMSE,
        pred,
        ref,
    ),
    "mean_sRMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.sRMSE,
        pred,
        ref,
    ),
    "mean_fourier_MAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_MAE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_fourier_nMAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_nMAE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_fourier_MSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_MSE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_fourier_nMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_nMSE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_fourier_RMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_RMSE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_fourier_nRMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.fourier_nRMSE,
        pred,
        ref,
        low=int(metric_config.split(";")[1]),
        high=int(metric_config.split(";")[2]),
        derivative_order=float(metric_config.split(";")[3]),
    ),
    "mean_H1_MAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_MAE,
        pred,
        ref,
    ),
    "mean_H1_nMAE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_nMAE,
        pred,
        ref,
    ),
    "mean_H1_MSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_MSE,
        pred,
        ref,
    ),
    "mean_H1_nMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_nMSE,
        pred,
        ref,
    ),
    "mean_H1_RMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_RMSE,
        pred,
        ref,
    ),
    "mean_H1_nRMSE": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.H1_nRMSE,
        pred,
        ref,
    ),
    "mean_correlation": lambda metric_config: lambda pred, ref: ex.metrics.mean_metric(
        ex.metrics.correlation,
        pred,
        ref,
    ),
}



from ._convection import Burgers, Convection, KuramotoSivashinskyConservative
from ._linear import (
    Advection,
    AdvectionDiffusion,
    Diffusion,
    Dispersion,
    FirstFour,
    HyperDiffusion,
    Linear,
    LinearSimple,
)
from ._nonlinear import (
    BurgersSingleChannel,
    FisherKPP,
    KortewegDeVries,
    KuramotoSivashinsky,
    Nonlinear,
)

scenario_dict = {
    "diff_lin": Linear,
    "diff_lin_simple": LinearSimple,
    "diff_adv": Advection,
    "diff_diff": Diffusion,
    "diff_adv_diff": AdvectionDiffusion,
    "diff_disp": Dispersion,
    "diff_hyp_diff": HyperDiffusion,
    "diff_four": FirstFour,
    "diff_conv": Convection,
    "diff_burgers": Burgers,
    "diff_kdv": KortewegDeVries,
    "diff_ks_cons": KuramotoSivashinskyConservative,
    "diff_ks": KuramotoSivashinsky,
    "diff_nonlin": Nonlinear,
    "diff_burgers_sc": BurgersSingleChannel,
    "diff_fisher": FisherKPP,
}



"""
Utilities to scrape APEBench datasets into numpy arrays and save them to disk.
"""

import json
import logging
from dataclasses import asdict

import jax.numpy as jnp

from .scenarios import scenario_dict


def scrape_data_and_metadata(
    folder: str = None,
    *,
    scenario: str,
    name: str = "auto",
    **scenario_kwargs,
):
    """
    Produce train data, test data, and metadata for a given scenario. Optionally
    write them to disk.

    **Arguments:**

    - `folder` (str, optional): Folder to save the data and metadata to. If
        None, returns the data and metadata as jax arrays and a dictionary,
        respectively.
    - `scenario` (str): Name of the scenario to produce data for. Must be one of
        `apebench.scenarios.scenario_dict`.
    - `name` (str, optional): Name of the scenario. If "auto", the name is
        automatically generated based on the scenario and its additional
        arguments.
    - `**scenario_kwargs`: Additional arguments to pass to the scenario. All
        attributes of a scenario can be modified by passing them as keyword
        arguments.
    """
    scenario = scenario_dict[scenario](**scenario_kwargs)
    if name == "auto":
        name = scenario.get_scenario_name()

        additional_infos = []
        for key, value in scenario_kwargs.items():
            additional_infos.append(f"{key}={value}")
        if len(additional_infos) > 0:
            additional_infos = ", ".join(additional_infos)
            additional_infos = "__" + additional_infos
        else:
            additional_infos = ""

        name += additional_infos

    logging.info(f"Producing train data for {name}")
    train_data = scenario.get_train_data()
    train_num_nans = jnp.sum(jnp.isnan(train_data))
    if train_num_nans > 0:
        logging.warning(f"Train data contains {train_num_nans} NaNs")

    logging.info(f"Producing test data for {name}")
    test_data = scenario.get_test_data()
    test_num_nans = jnp.sum(jnp.isnan(test_data))
    if test_num_nans > 0:
        logging.warning(f"Test data contains {test_num_nans} NaNs")

    info = asdict(scenario)

    metadata = {
        "name": name,
        "info": info,
    }

    if folder is not None:
        with open(f"{folder}/{name}.json", "w") as f:
            json.dump(metadata, f)
        jnp.save(f"{folder}/{name}_train.npy", train_data)
        jnp.save(f"{folder}/{name}_test.npy", test_data)

        del train_data, test_data
    else:
        return train_data, test_data, metadata


CURATION_APEBENCH_V1 = [
    # 1D - Linear
    {
        "scenario": "diff_adv",
    },
    {
        "scenario": "diff_diff",
    },
    {
        "scenario": "diff_adv_diff",
    },
    {
        "scenario": "diff_disp",
    },
    {
        "scenario": "diff_hyp_diff",
    },
    # 1D - Nonlinear
    {"scenario": "diff_burgers"},
    {"scenario": "diff_kdv"},
    {"scenario": "diff_ks"},
    {"scenario": "diff_ks_cons"},
    # 1D - Reaction-Diffusion
    {"scenario": "diff_fisher"},
    # 2D - Linear
    {"scenario": "diff_adv", "num_spatial_dims": 2},
    {"scenario": "diff_diff", "num_spatial_dims": 2},
    {"scenario": "diff_adv_diff", "num_spatial_dims": 2},
    {"scenario": "diff_disp", "num_spatial_dims": 2},
    {"scenario": "diff_hyp_diff", "num_spatial_dims": 2},
    # 2D - Linear Special
    {
        "scenario": "phy_unbal_adv",
        "num_spatial_dims": 2,
        "advection_coef_vector": (0.01, -0.04),
    },
    {"scenario": "phy_diag_diff", "num_spatial_dims": 2},
    {"scenario": "phy_aniso_diff", "num_spatial_dims": 2},
    {"scenario": "phy_mix_disp", "num_spatial_dims": 2},
    {"scenario": "phy_mix_hyp", "num_spatial_dims": 2},
    # 2D - Nonlinear
    {"scenario": "diff_burgers", "num_spatial_dims": 2},
    {"scenario": "diff_burgers_sc", "num_spatial_dims": 2},
    {"scenario": "diff_kdv", "num_spatial_dims": 2},
    {"scenario": "diff_ks", "num_spatial_dims": 2},
    {"scenario": "phy_decay_turb", "num_spatial_dims": 2},
    {"scenario": "phy_kolm_flow", "num_spatial_dims": 2},
    # 2D - Reaction-Diffusion
    {"scenario": "diff_fisher", "num_spatial_dims": 2},
    {"scenario": "phy_gs_type", "num_spatial_dims": 2},
    {"scenario": "phy_sh", "num_spatial_dims": 2},
    # 3D - Linear
    {
        "scenario": "diff_adv",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_diff",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_adv_diff",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_disp",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_hyp_diff",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    # 3D - Linear Special
    {
        "scenario": "phy_unbal_adv",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_diag_diff",
        "num_spatial_dims": 3,
        "num_points": 32,
        "diffusion_coef_vector": (0.001, 0.002, 0.0004),
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_aniso_diff",
        "num_spatial_dims": 3,
        "num_points": 32,
        "diffusion_coef_matrix": (
            (0.001, 0.0005, 0.0003),
            (0.0005, 0.002, 0.0002),
            (0.0003, 0.0002, 0.0004),
        ),
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_mix_disp",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_mix_hyp",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    # 3D - Nonlinear
    {
        "scenario": "diff_burgers",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_burgers_sc",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_kdv",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "diff_ks",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    # 3D - Reaction-Diffusion
    {
        "scenario": "diff_fisher",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_gs_type",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
    {
        "scenario": "phy_sh",
        "num_spatial_dims": 3,
        "num_points": 32,
        "num_test_samples": 10,
    },
]
"""
Collection of default scenarios as used in the original APEBench paper
"""



from ._convection import Burgers, Convection, KuramotoSivashinskyConservative
from ._gray_scott import GrayScott, GrayScottType
from ._linear import (
    Advection,
    AdvectionDiffusion,
    Diffusion,
    Dispersion,
    FirstFour,
    HyperDiffusion,
    Linear,
    LinearSimple,
)
from ._navier_stokes import DecayingTurbulence, KolmogorovFlow
from ._nonlinear import (
    BurgersSingleChannel,
    KortewegDeVries,
    KuramotoSivashinsky,
    Nonlinear,
)
from ._poisson import Poisson
from ._polynomial import FisherKPP, Polynomial
from ._special_linear import (
    AnisotropicDiffusion,
    DiagonalDiffusion,
    SpatiallyMixedDispersion,
    SpatiallyMixedHyperDiffusion,
    UnbalancedAdvection,
)
from ._swift_hohenberg import SwiftHohenberg

scenario_dict = {
    "phy_poisson": Poisson,
    "phy_sh": SwiftHohenberg,
    "phy_gs": GrayScott,
    "phy_gs_type": GrayScottType,
    "phy_decay_turb": DecayingTurbulence,
    "phy_kolm_flow": KolmogorovFlow,
    "phy_lin": Linear,
    "phy_lin_simple": LinearSimple,
    "phy_adv": Advection,
    "phy_diff": Diffusion,
    "phy_adv_diff": AdvectionDiffusion,
    "phy_disp": Dispersion,
    "phy_hyp_diff": HyperDiffusion,
    "phy_four": FirstFour,
    "phy_nonlin": Nonlinear,
    "phy_burgers_sc": BurgersSingleChannel,
    "phy_kdv": KortewegDeVries,
    "phy_ks": KuramotoSivashinsky,
    "phy_conv": Convection,
    "phy_burgers": Burgers,
    "phy_ks_cons": KuramotoSivashinskyConservative,
    "phy_poly": Polynomial,
    "phy_fisher": FisherKPP,
    "phy_unbal_adv": UnbalancedAdvection,
    "phy_diag_diff": DiagonalDiffusion,
    "phy_aniso_diff": AnisotropicDiffusion,
    "phy_mix_disp": SpatiallyMixedDispersion,
    "phy_mix_hyp": SpatiallyMixedHyperDiffusion,
}



from typing import Callable, Dict

import equinox as eqx
import pdequinox as pdeqx
from jaxtyping import PRNGKeyArray


def dil_constructor(
    architecture_config: str,
    num_spatial_dims: int,
    num_points: int,
    num_channels: int,
    activation_fn: Callable,
    key: PRNGKeyArray,
):
    architecture_args = architecture_config.split(";")
    dilation_depth = int(architecture_args[1])
    hidden_channels = int(architecture_args[2])
    num_blocks = int(architecture_args[3])

    dilation_rates = [2**i for i in range(dilation_depth + 1)]
    dilation_rates = dilation_rates + dilation_rates[::-1][1:]

    return pdeqx.arch.DilatedResNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=hidden_channels,
        num_blocks=num_blocks,
        dilation_rates=dilation_rates,
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    )


architecture_dict: Dict[
    str,  # architecture_name
    Callable[
        [
            str,  # architecture_config
            int,  # num_spatial_dims
            int,  # num_points
            int,  # num_channels
            Callable,  # activation_fn
            PRNGKeyArray,  # key
        ],
        eqx.Module,  # architecture
    ],
] = {
    "conv": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ConvNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=int(architecture_config.split(";")[1]),
        depth=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
    "res": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ClassicResNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=int(architecture_config.split(";")[1]),
        num_blocks=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
    "unet": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ClassicUNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=int(architecture_config.split(";")[1]),
        num_levels=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
    "dil": dil_constructor,
    "fno": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ClassicFNO(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        num_modes=int(architecture_config.split(";")[1]),
        hidden_channels=int(architecture_config.split(";")[2]),
        num_blocks=int(architecture_config.split(";")[3]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
    "mlp": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.MLP(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        num_points=num_points,
        width_size=int(architecture_config.split(";")[1]),
        depth=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        key=key,
    ),
    "pure": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.conv.PhysicsConv(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        kernel_size=int(architecture_config.split(";")[1]),
        use_bias=False,  # !!! no bias
        boundary_mode="periodic",
        key=key,
    ),
    "mores": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ModernResNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=int(architecture_config.split(";")[1]),
        num_blocks=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
    "mounet": lambda architecture_config, num_spatial_dims, num_points, num_channels, activation_fn, key: pdeqx.arch.ModernUNet(
        num_spatial_dims=num_spatial_dims,
        in_channels=num_channels,
        out_channels=num_channels,
        hidden_channels=int(architecture_config.split(";")[1]),
        num_levels=int(architecture_config.split(";")[2]),
        activation=activation_fn,
        boundary_mode="periodic",
        key=key,
    ),
}



import equinox as eqx
import jax


class CorrectedStepper(eqx.Module):
    """
    Take a given coarse stepper (whose parameters are fixed) and a network that
    is supposed to correct it.
    """

    coarse_stepper: eqx.Module
    network: eqx.Module
    mode: str

    def __call__(
        self,
        x,
    ):
        # Detach the coarse solver because it is not supposed to be trained
        coarse_solver_detached = jax.lax.stop_gradient(self.coarse_stepper)

        if self.mode == "parallel":
            network_prediction = self.network(x)

            coarse_solver_prediction = coarse_solver_detached(x)

            next_x = coarse_solver_prediction + network_prediction
        elif self.mode == "sequential":
            coarse_solver_prediction = coarse_solver_detached(x)

            network_prediction = self.network(coarse_solver_prediction)

            next_x = network_prediction
        elif self.mode == "sequential_with_bypass":
            coarse_solver_prediction = coarse_solver_detached(x)

            network_prediction = self.network(coarse_solver_prediction)

            next_x = coarse_solver_prediction + network_prediction
        else:
            raise ValueError(f"Unknown mode {self.mode}")

        return next_x



from typing import Literal

import exponax as ex
from exponax.ic import BaseRandomICGenerator

from ..._base_scenario import BaseScenario


class GrayScott(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 10.0
    num_channels: int = 2  # Overwrite

    num_substeps: int = 10

    ic_config: str = "gray_scott_blobs"  # Overwrite

    feed_rate: float = 0.04
    kill_rate: float = 0.06
    diffusivity_1: float = 2e-5
    diffusivity_2: float = 1e-5

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError("GrayScott is only supported for 2D and 3D")
        if self.num_channels != 2:
            raise ValueError("Number of channels must be 2 for GrayScott")

    def get_ic_generator(self) -> BaseRandomICGenerator:
        return ex.ic.RandomMultiChannelICGenerator(
            [
                ex.ic.RandomGaussianBlobs(self.num_spatial_dims, one_complement=True),
                ex.ic.RandomGaussianBlobs(self.num_spatial_dims),
            ]
        )

    def _build_stepper(self, dt):
        substepped_stepper = ex.stepper.reaction.GrayScott(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=dt / self.num_substeps,
            feed_rate=self.feed_rate,
            kill_rate=self.kill_rate,
            diffusivity_1=self.diffusivity_1,
            diffusivity_2=self.diffusivity_2,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self):
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_gs"


# Following scenario follows
# "https://www.ljll.fr/hecht/ftp/ff++/2015-cimpa-IIT/edp-tuto/Pearson.pdf"


class GrayScottType(BaseScenario):
    domain_extent: float = 2.5
    dt: float = 20.0
    num_channels: int = 2  # Overwrite

    coarse_proportion: float = 0.5

    num_substeps: int = 20

    ic_config: str = "gray_scott_blobs"  # Overwrite

    diffusivity_1: float = 2e-5
    diffusivity_2: float = 1e-5
    pattern_type: Literal[
        "alpha",
        "beta",
        "gamma",
        "delta",
        "epsilon",
        "zeta",
        "eta",
        "theta",
        "iota",
        "kappa",
        "lambda",
        "mu",
    ] = "theta"

    def __post_init__(self):
        if not (self.num_spatial_dims in [2, 3]):
            raise ValueError(
                "Only 2 and 3 spatial dimensions are supported for GrayScott"
            )
        if self.num_channels != 2:
            raise ValueError("Number of channels must be 2 for GrayScott")

    def get_feed_and_kill_rate(
        self,
        pattern_type: Literal[
            "alpha",
            "beta",
            "gamma",
            "delta",
            "epsilon",
            "zeta",
            "eta",
            "theta",
            "iota",
            "kappa",
            "lambda",
            "mu",
        ],
    ):
        if pattern_type == "alpha":
            feed_rate = 0.008
            kill_rate = 0.046
        elif pattern_type == "beta":
            feed_rate = 0.020
            kill_rate = 0.046
        elif pattern_type == "gamma":
            feed_rate = 0.024
            kill_rate = 0.056
        elif pattern_type == "delta":
            feed_rate = 0.028
            kill_rate = 0.056
        elif pattern_type == "epsilon":
            feed_rate = 0.02
            kill_rate = 0.056
        elif pattern_type == "zeta":
            # Does not seem to work
            feed_rate = 0.024
            kill_rate = 0.06
        elif pattern_type == "eta":
            # Does not seem to work
            feed_rate = 0.036
            kill_rate = 0.063
        elif pattern_type == "theta":
            feed_rate = 0.04
            kill_rate = 0.06
        elif pattern_type == "iota":
            feed_rate = 0.05
            kill_rate = 0.0605
        elif pattern_type == "kappa":
            feed_rate = 0.052
            kill_rate = 0.063
        elif pattern_type == "lambda":
            # Does not seem to work
            feed_rate = 0.036
            kill_rate = 0.0655
        elif pattern_type == "mu":
            # Does not seem to work
            feed_rate = 0.044
            kill_rate = 0.066
        else:
            raise ValueError("Invalid pattern type")

        return feed_rate, kill_rate

    def get_ic_generator(self) -> BaseRandomICGenerator:
        return ex.ic.RandomMultiChannelICGenerator(
            [
                ex.ic.RandomGaussianBlobs(self.num_spatial_dims, one_complement=True),
                ex.ic.RandomGaussianBlobs(self.num_spatial_dims),
            ]
        )

    def _build_stepper(self, dt):
        feed_rate, kill_rate = self.get_feed_and_kill_rate(self.pattern_type)
        substepped_stepper = ex.stepper.reaction.GrayScott(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=dt / self.num_substeps,
            feed_rate=feed_rate,
            kill_rate=kill_rate,
            diffusivity_1=self.diffusivity_1,
            diffusivity_2=self.diffusivity_2,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self) -> ex.BaseStepper:
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_gs_{self.pattern_type}"



from math import pi

import exponax as ex

from ..._base_scenario import BaseScenario


class SwiftHohenberg(BaseScenario):
    domain_extent: float = 10.0 * pi
    dt: float = 0.1

    num_substeps: int = 5

    reactivity: float = 0.7
    critical_number: float = 1.0
    polynomial_coefficients: tuple[float, ...] = (0.0, 0.0, 1.0, -1.0)

    coarse_proportion: float = 0.5

    def __post_init__(self):
        if self.num_spatial_dims == 1:
            raise ValueError("Swift-Hohenberg is only supported for 2D and 3D")

    def _build_stepper(self, dt):
        substepped_stepper = ex.stepper.reaction.SwiftHohenberg(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=dt / self.num_substeps,
            reactivity=self.reactivity,
            critical_number=self.critical_number,
            polynomial_coefficients=self.polynomial_coefficients,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self):
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_sh"



from ._convection import Burgers, Convection, KuramotoSivashinskyConservative
from ._linear import (
    Advection,
    AdvectionDiffusion,
    Diffusion,
    Dispersion,
    FirstFour,
    HyperDiffusion,
    Linear,
    LinearSimple,
)
from ._nonlinear import (
    BurgersSingleChannel,
    FisherKPP,
    KortewegDeVries,
    KuramotoSivashinsky,
    Nonlinear,
)

scenario_dict = {
    "norm_lin": Linear,
    "norm_adv": Advection,
    "norm_diff": Diffusion,
    "norm_adv_diff": AdvectionDiffusion,
    "norm_disp": Dispersion,
    "norm_fisher": FisherKPP,
    "norm_four": FirstFour,
    "norm_hypdiff": HyperDiffusion,
    "norm_nonlin": Nonlinear,
    "norm_conv": Convection,
    "norm_burgers": Burgers,
    "norm_kdv": KortewegDeVries,
    "norm_ks_cons": KuramotoSivashinskyConservative,
    "norm_ks": KuramotoSivashinsky,
    "norm_burgers_sc": BurgersSingleChannel,
    "norm_lin_simple": LinearSimple,
}



import importlib.metadata

import exponax
import pdequinox
import trainax

from . import _scraper as scraper
from . import components, scenarios
from ._base_scenario import BaseScenario
from ._run import (
    get_experiment_name,
    melt_concat_from_list,
    melt_concat_loss_from_list,
    melt_concat_metrics_from_list,
    melt_concat_sample_rollouts_from_list,
    run_experiment,
    run_study,
    run_study_convenience,
)
from ._utils import (
    aggregate_gmean,
    check_for_nan,
    melt_data,
    melt_loss,
    melt_metrics,
    melt_sample_rollouts,
    read_in_kwargs,
    relative_by_config,
    split_train,
)

__version__ = importlib.metadata.version("apebench")

__all__ = [
    "exponax",
    "pdequinox",
    "get_experiment_name",
    "melt_concat_from_list",
    "melt_concat_loss_from_list",
    "melt_concat_metrics_from_list",
    "melt_concat_sample_rollouts_from_list",
    "run_experiment",
    "run_study_convenience",
    "run_study",
    "scenarios",
    "trainax",
    "melt_data",
    "melt_loss",
    "melt_metrics",
    "melt_sample_rollouts",
    "read_in_kwargs",
    "arch_extensions",
    "scraper",
    "BaseScenario",
    "aggregate_gmean",
    "relative_by_config",
    "split_train",
    "components",
    "check_for_nan",
]



from . import difficulty, normalized, physical

scenario_dict = {
    **normalized.scenario_dict,
    **physical.scenario_dict,
    **difficulty.scenario_dict,
}

guaranteed_non_nan: list[
    tuple[
        str,  # scenario_name
        int,  # num_spatial_dims
    ]
] = [
    # Normalized Interface
    ("norm_lin", 1),
    ("norm_adv", 1),
    ("norm_diff", 1),
    ("norm_adv_diff", 1),
    ("norm_disp", 1),
    ("norm_fisher", 1),
    ("norm_four", 1),
    ("norm_hypdiff", 1),
    ("norm_nonlin", 1),
    ("norm_conv", 1),
    ("norm_burgers", 1),
    ("norm_kdv", 1),
    ("norm_ks_cons", 1),
    ("norm_ks", 1),
    ("norm_burgers_sc", 1),
    ("norm_lin_simple", 1),
    # Physical Interface
    ("phy_poisson", 1),
    ("phy_sh", 2),
    ("phy_gs", 2),
    ("phy_gs_type", 2),
    ("phy_decay_turb", 2),
    ("phy_kolm_flow", 2),
    ("phy_lin", 1),
    ("phy_lin_simple", 1),
    ("phy_adv", 1),
    ("phy_diff", 1),
    ("phy_adv_diff", 1),
    ("phy_disp", 1),
    ("phy_hyp_diff", 1),
    ("phy_four", 1),
    ("phy_nonlin", 1),
    ("phy_burgers_sc", 1),
    ("phy_kdv", 1),
    ("phy_ks", 1),
    ("phy_conv", 1),
    ("phy_burgers", 1),
    ("phy_ks_cons", 1),
    ("phy_poly", 1),
    ("phy_fisher", 1),
    ("phy_unbal_adv", 3),
    ("phy_diag_diff", 2),
    ("phy_aniso_diff", 2),
    ("phy_mix_disp", 2),
    ("phy_mix_hyp", 2),
    # In the difficulty-based interface all scenarios are guaranteed to be
    # non-nan for 1d, 2d, and 3d
]



from typing import Callable, Dict

import optax

optimizer_dict: Dict[
    str,  # optimizer_name
    Callable[
        [
            str,
        ],  # optim_config
        Callable[
            [
                optax.Schedule,
            ],  # lr_schedule
            optax.GradientTransformation,  # optimizer
        ],
    ],
] = {"adam": lambda optim_config: optax.adam}

lr_scheduler_dict: Dict[
    str,  # lr_scheduler_name
    Callable[
        [
            str,  # lr_scheduler_config
            int,  # num_training_steps
        ],
        optax.Schedule,  # lr_schedule
    ],
] = {
    "constant": lambda lr_scheduler_config, num_training_steps: optax.constant_schedule(
        float(lr_scheduler_config.split(";")[1]),
    ),
    "exp": lambda lr_scheduler_config, num_training_steps: optax.exponential_decay(
        init_value=float(lr_scheduler_config.split(";")[1]),
        transition_steps=int(lr_scheduler_config.split(";")[2]),
        decay_rate=float(lr_scheduler_config.split(";")[3]),
        staircase=lr_scheduler_config.split(";")[4] == "True",
    ),
    "warmup_cosine": lambda lr_scheduler_config, num_training_steps: optax.warmup_cosine_decay_schedule(
        init_value=float(lr_scheduler_config.split(";")[1]),
        peak_value=float(lr_scheduler_config.split(";")[2]),
        warmup_steps=int(lr_scheduler_config.split(";")[3]),
        decay_steps=num_training_steps,
    ),
}



import exponax as ex

from ..._base_scenario import BaseScenario


class Polynomial(BaseScenario):
    domain_extent: float = 1.0
    dt: float = 0.001

    a_coefs: tuple[float, ...] = (0.02, 0.0, 4e-6, 0.0, 0.0)
    poly_coefs: tuple[float, ...] = (0.0, 0.0, -0.02)

    num_substeps: int = 1

    coarse_proportion: float = 0.5

    order: int = 2
    dealiasing_fraction: float = 2 / 3
    num_circle_points: int = 16
    circle_radius: float = 1.0

    def __post_init__(self):
        pass

    def _build_stepper(self, dt):
        substepped_stepper = ex.stepper.generic.GeneralPolynomialStepper(
            num_spatial_dims=self.num_spatial_dims,
            domain_extent=self.domain_extent,
            num_points=self.num_points,
            dt=dt / self.num_substeps,
            linear_coefficients=self.a_coefs,
            polynomial_coefficients=self.poly_coefs,
            order=self.order,
            dealiasing_fraction=self.dealiasing_fraction,
            num_circle_points=self.num_circle_points,
            circle_radius=self.circle_radius,
        )

        if self.num_substeps == 1:
            stepper = substepped_stepper
        else:
            stepper = ex.RepeatedStepper(substepped_stepper, self.num_substeps)

        return stepper

    def get_ref_stepper(self):
        return self._build_stepper(self.dt)

    def get_coarse_stepper(self):
        return self._build_stepper(self.dt * self.coarse_proportion)

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_poly"


class FisherKPP(Polynomial):
    drag_coef: float = 20.0
    diffusion_coef: float = 0.004
    quadratic_coef: float = -20.0

    ic_config: str = "clamp;0.0;1.0;fourier;5;false;false"  # Overwrite

    def __post_init__(self):
        self.a_coefs = (self.drag_coef, 0.0, self.diffusion_coef, 0.0, 0.0)
        self.poly_coefs = (0.0, 0.0, self.quadratic_coef)

        super().__post_init__()

    def get_scenario_name(self) -> str:
        return f"{self.num_spatial_dims}d_phy_fisher"
